# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for defining methods that handle everything that
is related to versioning and dependencies of software.
"""
import logging
from typing import Dict, List, Tuple

import git
import pkg_resources
from git import InvalidGitRepositoryError

logger = logging.getLogger(__name__)


def get_installed_package_versions() -> Dict[str, str]:
    """
    Get installed python packages as name/version tuples.
    """
    return {d.project_name: d.version for d in pkg_resources.working_set}


def get_installed_package_list() -> List[str]:
    """
    Get installed python packages.
    """
    return ["{}=={}".format(d.project_name, d.version) for d in pkg_resources.working_set]


def get_installed_pip_package_versions() -> List[str]:
    """
    Get versions of installed pip packages.
    Deprecated: Use get_installed_package_list
    """
    logger.warning(
        "get_installed_pip_package_versions() is a deprecated alias for get_installed_package_list()"
    )
    return get_installed_package_list()


def get_git_info() -> Tuple[str, str, str]:
    """
    Get info about the current git repository

    Returns:
        Tuple consisting of
        - repository identifier
        - sha as string
        - branch-name as string
    """
    try:
        repo = git.Repo(search_parent_directories=True)
        sha: str = str(repo.head.object.hexsha)

        try:
            branch: str = str(repo.active_branch)
        except TypeError as type_error:
            branch = str(type_error)

        return str(repo), sha, branch
    except InvalidGitRepositoryError:
        logger.warning("Could not find a valid git repository for SCM info")
        return "non-scm", "", ""


def write_pip_info_to_file(output_requirements_path: str) -> None:
    """
    Args:
        output_requirements_path: Path to the requirements file to generate
    """
    with open(file=output_requirements_path, mode="w") as pip_package_file:
        pip_packages = get_installed_pip_package_versions()
        for pip_package in pip_packages:
            pip_package_file.write(f"{pip_package}\n")
        logger.info(
            "Wrote currently installed package version to requirements file: %s",
            pip_package_file,
        )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations regarding time and json operations"""
import inspect
import json
from datetime import datetime, timedelta
from typing import Any


def get_current_timestamp(format_string: str = "%d-%b-%Y (%H:%M:%S.%f)") -> str:
    """
    Returns the current time in the given format
    Args:
        format_string: String, defining the format of the timestamp

    Returns: String, the formatted timestamp

    """

    timestamp = datetime.now()

    timestamp_string = timestamp.strftime(format_string)

    return timestamp_string


def timestamp_to_datetime(datetime_object: datetime) -> timedelta:
    """
    Transforms a timestamp to a timedelta based on the given object's attributes

    Args:
        datetime_object: datetime, a timestamp

    Returns: a timedelta object

    """

    return timedelta(
        hours=datetime_object.hour,
        minutes=datetime_object.minute,
        seconds=datetime_object.second,
        microseconds=datetime_object.microsecond,
    )


class CustomJSONEncoder(json.JSONEncoder):
    """Class for encoding json objects"""

    def default(self, obj: Any) -> Any:
        """
        Sets the class's default attribute to the json version of the given object

        Args:
            obj: Any, an object that should be transformed to json format

        Returns: Any, the given object without changes

        """

        if hasattr(obj, "to_json"):
            return self.default(obj.to_json())
        elif hasattr(obj, "__dict__"):
            d = dict(
                (key, value)
                for key, value in inspect.getmembers(obj)
                if not key.startswith("__")
                and not inspect.isabstract(value)
                and not inspect.isbuiltin(value)
                and not inspect.isfunction(value)
                and not inspect.isgenerator(value)
                and not inspect.isgeneratorfunction(value)
                and not inspect.ismethod(value)
                and not inspect.ismethoddescriptor(value)
                and not inspect.isroutine(value)
            )
            return self.default(d)
        return obj



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for different utility operations during data preparation
"""

import logging
import os
import xml.dom.minidom as mmd
import xml.etree.ElementTree as ET_xml
from typing import List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.utils.file_utils import ensure_dir
from mlcvzoo_base.utils.implicit_path_replacements import ImplicitReplacement
from mlcvzoo_base.utils.xml_utils import create_bbox_xml_entry, create_xml_root

logger = logging.getLogger(__name__)


def replace_index_by_dir(annotation: BaseAnnotation) -> BaseAnnotation:
    """
    Replaces replacement_string parts in image_path and annotation_path attributes of
    an annotation by respective directory information (image_dir and annotation_dir)

    Args:
        annotation: BaseAnnotation object

    Returns: BaseAnnotation, the input object with different path attributes

    """

    if annotation.replacement_string is not None:
        # TODO: use directory replacement
        annotation.image_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.image_path,
            value=annotation.replacement_string,
            replacement_value=annotation.image_dir,
            how=ImplicitReplacement.FIRST,
        )

        # TODO: use directory replacement
        annotation.annotation_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.annotation_path,
            value=annotation.replacement_string,
            replacement_value=annotation.annotation_dir,
            how=ImplicitReplacement.FIRST,
        )

        return ensure_abspath(annotation=annotation)


def replace_dir_by_index(annotation: BaseAnnotation) -> BaseAnnotation:
    """
    Replaces directory parts (image_dir and annotation_dir) in image_path and
    annotation_path attributes of an annotation by replacement_string information

    Args:
        annotation: BaseAnnotation object

    Returns: BaseAnnotation, the input object with different path attributes

    """

    if annotation.replacement_string is not None:
        annotation.image_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.image_path,
            value=annotation.image_dir,
            replacement_value=annotation.replacement_string,
            how=ImplicitReplacement.FIRST,
        )

        annotation.annotation_path = ImplicitReplacement.replace_string_in_path(
            file_path=annotation.annotation_path,
            value=annotation.annotation_dir,
            replacement_value=annotation.replacement_string,
            how=ImplicitReplacement.FIRST,
        )

    return annotation


def ensure_abspath(annotation: BaseAnnotation) -> BaseAnnotation:
    """
    Sets the given annotation's path and directory attributes to absolute paths

    Args:
        annotation: BaseAnnotation object

    Returns: BaseAnnotation, the input object with different path attributes

    """

    annotation.image_path = os.path.abspath(annotation.image_path)
    annotation.image_dir = os.path.abspath(annotation.image_dir)
    annotation.annotation_path = os.path.abspath(annotation.annotation_path)
    annotation.annotation_dir = os.path.abspath(annotation.annotation_dir)

    return annotation


def annotation_to_xml(
    annotation: BaseAnnotation,
    allowed_classes: Optional[List[str]] = None,
) -> None:
    """
    Transforms the given annotation object to XML format and saves it to
    a file at annotation_path location

    Args:
        annotation: BaseAnnotation object
        allowed_classes: Optional list of strings (allowed classes)

    Returns: None

    """

    root = create_xml_root(
        file_path=annotation.image_path,
        width=annotation.image_shape[0],
        height=annotation.image_shape[1],
    )

    assert annotation.image_shape[0] > 0
    assert annotation.image_shape[1] > 0

    if len(annotation.get_bounding_boxes(include_segmentations=True)) == 0:
        logger.warning("annotation has no bounding_boxes!")

    for bounding_box in annotation.get_bounding_boxes(include_segmentations=True):
        if allowed_classes is not None and bounding_box.class_name not in allowed_classes:
            continue

        root = create_bbox_xml_entry(
            root,
            bounding_box.class_name,
            bounding_box.box.xmin,
            bounding_box.box.ymin,
            bounding_box.box.xmax,
            bounding_box.box.ymax,
        )

    logger.info("Write xml file: %s", annotation.annotation_path)

    ensure_dir(file_path=annotation.annotation_path, verbose=True)

    tree_string = "".join(ET_xml.tostring(element=root).decode("utf-8").split())
    pretty_tree_string = mmd.parseString(tree_string).toprettyxml(
        indent="  ",
    )

    with open(file=annotation.annotation_path, mode="w", encoding="utf8") as annotation_file:
        annotation_file.write(pretty_tree_string)


def save_bbox_snippets(annotation: BaseAnnotation, output_dir: str) -> None:
    raise NotImplementedError("Currently not implemented!")
    # TODO:
    # image_path = self.image_path
    #
    # snippet_base_dir = os.path.join(
    #     output_dir,
    #     "snippets"
    # )
    #
    # for index, d in enumerate(self.bounding_boxes):
    #     snippet_name = os.path.basename(image_path)[:-4] + "_" + \
    #                    d.class_name + "_" + str(index) + os.path.basename(image_path)[-4:]
    #
    #     snippet_dir = os.path.join(snippet_base_dir, d.class_name)
    #
    #     if not os.path.exists(snippet_dir):
    #         os.makedirs(snippet_dir)
    #
    #         print("Create directory: {}".format(snippet_dir))
    #
    #     snippet_path = os.path.join(snippet_dir, snippet_name)
    #
    #     if not os.path.isfile(snippet_path):
    #         img = cv2.imread(image_path)
    #
    #         crop_img = img[d.box.ymin:d.box.ymax, d.box.xmin:d.box.xmax]
    #
    #         # NOTE: comment in to show image
    #         # cv2.imshow(snippet_name, crop_img)
    #         # cv2.waitKey(0)
    #
    #         print("Write snippet: {}".format(snippet_path))
    #         cv2.imwrite(snippet_path, crop_img)


def filter_annotation(
    annotation: BaseAnnotation,
    class_ids: Optional[List[int]] = None,
    class_names: Optional[List[str]] = None,
    classification_score: float = 0.0,
    bounding_box_score: float = 0.0,
    segmentation_score: float = 0.0,
) -> BaseAnnotation:
    """
    Filter a given annotation object. Afterwards the object only contains
    the classifications, bounding-boxes and segmentations that fulfill:

    - the class id matches one of the given ids given by 'class_ids' (only when the list is
      not None nor empty)
    - the class name matches one of the given names given by 'class_names' (only when the list is
      not None nor empty)
    - the score is higher than the one associated with this datastructure

    Args:
        annotation: the annotation object to filter
        class_ids: the class ids that are valid
        class_names: the class names that are valid
        classification_score: the score a any Classification object has to fulfill
        bounding_box_score: the score a any BoundingBox object has to fulfill
        segmentation_score: the score a any Segmentation object has to fulfill

    Returns:
        The filter annotation object
    """

    filtered_annotation: BaseAnnotation = BaseAnnotation(
        image_path=annotation.image_path,
        annotation_path=annotation.annotation_path,
        image_shape=annotation.image_shape,
        image_dir=annotation.image_dir,
        replacement_string=annotation.replacement_string,
        annotation_dir=annotation.annotation_dir,
        classifications=[],
        bounding_boxes=[],
        segmentations=[],
    )

    for classification in annotation.classifications:
        if (
            class_ids is not None
            and len(class_ids) > 0
            and classification.class_id not in class_ids
        ) or (
            class_names is not None
            and len(class_names) > 0
            and classification.class_name not in class_names
        ):
            continue

        if classification.score >= classification_score:
            filtered_annotation.classifications.append(classification)

    for bounding_box in annotation.bounding_boxes:
        if (
            class_ids is not None and len(class_ids) > 0 and bounding_box.class_id not in class_ids
        ) or (
            class_names is not None
            and len(class_names) > 0
            and bounding_box.class_name not in class_names
        ):
            continue

        if bounding_box.score >= bounding_box_score:
            filtered_annotation.bounding_boxes.append(bounding_box)

    for segmentation in annotation.segmentations:
        if (
            class_ids is not None and len(class_ids) > 0 and segmentation.class_id not in class_ids
        ) or (
            class_names is not None
            and len(class_names) > 0
            and segmentation.class_name not in class_names
        ):
            continue

        if segmentation.score >= segmentation_score:
            filtered_annotation.segmentations.append(segmentation)

    return filtered_annotation


def filter_annotations(
    annotations: List[BaseAnnotation],
    class_ids: Optional[List[int]] = None,
    class_names: Optional[List[str]] = None,
    classification_score: float = 0.0,
    bounding_box_score: float = 0.0,
    segmentation_score: float = 0.0,
) -> List[BaseAnnotation]:
    """
    Filter a given list of annotations. Afterwards each annotation only contains
    the classifications, bounding-boxes and segmentations that fulfill:

    - the class id matches one of the given ids given by 'class_ids' (only when the list is
      not None nor empty)
    - the class name matches one of the given names given by 'class_names' (only when the list is
      not None nor empty)
    - the score is higher than the one associated with this datastructure

    Args:
        annotations: the list of annotations to filter
        class_ids: the class ids that are valid
        class_names: the class names that are valid
        classification_score: the score a any Classification object has to fulfill
        bounding_box_score: the score a any BoundingBox object has to fulfill
        segmentation_score: the score a any Segmentation object has to fulfill

    Returns:
        The filtered list of annotations
    """

    filtered_annotations: List[BaseAnnotation] = []

    for annotation in annotations:
        filtered_annotations.append(
            filter_annotation(
                annotation=annotation,
                class_ids=class_ids,
                class_names=class_names,
                classification_score=classification_score,
                bounding_box_score=bounding_box_score,
                segmentation_score=segmentation_score,
            )
        )

    return filtered_annotations



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

""" Module for handling annotations according to configuration"""

import logging
import xml.etree.ElementTree as ET_xml
from typing import Any, Dict, List, Optional, Tuple, cast

from config_builder import ConfigBuilder

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.dataset_info import BaseDatasetInfo
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.configuration.annotation_handler_config import AnnotationHandlerConfig
from mlcvzoo_base.data_preparation.annotation_parser.coco_annotation_parser import (
    COCOAnnotationParser,
)
from mlcvzoo_base.data_preparation.annotation_parser.csv_annotation_parser import (
    CSVAnnotationParser,
)
from mlcvzoo_base.data_preparation.annotation_parser.cvat_annotation_parser import (
    CVATAnnotationParser,
)
from mlcvzoo_base.data_preparation.annotation_parser.label_studio_annotation_parser import (
    LabelStudioAnnotationParser,
)
from mlcvzoo_base.data_preparation.annotation_parser.mot_annotation_parser import (
    MOTAnnotationParser,
)
from mlcvzoo_base.data_preparation.annotation_parser.pascal_voc_annotation_parser import (
    PascalVOCAnnotationParser,
)
from mlcvzoo_base.data_preparation.annotation_writer.csv_annotation_writer import (
    CSVAnnotationWriter,
)
from mlcvzoo_base.data_preparation.annotation_writer.darknet_annotation_writer import (
    DarknetAnnotationWriter,
)
from mlcvzoo_base.data_preparation.structs import CSVOutputStringFormats

logger = logging.getLogger(__name__)


class AnnotationHandler:
    """Class for handling annotations"""

    csv_directory_replacement_string = "IMAGE_DIR_{}"
    csv_base_file_name: str

    def __init__(
        self,
        configuration: Optional[AnnotationHandlerConfig] = None,
        yaml_config_path: Optional[str] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        mapper: Optional[AnnotationClassMapper] = None,
    ) -> None:
        """

        Args:
            configuration: AnnotationHandlerConfig object
            yaml_config_path: path to configuration in yaml format
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            mapper: The mapper object that states the mapping of classes
        """

        if configuration is None:
            self.configuration: AnnotationHandlerConfig = self.create_configuration(
                from_yaml=yaml_config_path,
                configuration=configuration,
                string_replacement_map=string_replacement_map,
            )
        else:
            self.configuration = configuration

        if mapper is None:
            if self.configuration.class_mapping is not None:
                self.mapper = AnnotationClassMapper(
                    class_mapping=self.configuration.class_mapping,
                    reduction_mapping=self.configuration.reduction_class_mapping,
                )
            else:
                raise ValueError(
                    "Neither a mapper nor a class-mapping in the configuration is given!"
                )
        else:
            self.mapper = mapper

        self.train_info_dict: Dict[str, Any] = {}
        self.eval_info_dict: Dict[str, Any] = {}

        self.list_splits: List[Tuple[List[str], BaseDatasetInfo]] = []

        self.replace_data_dirs: Dict[str, str] = dict()

    @staticmethod
    def create_configuration(
        from_yaml: Optional[str] = None,
        configuration: Optional[AnnotationHandlerConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> AnnotationHandlerConfig:
        try:
            logger.info(
                "\nBuild AnnotationHandlerConfig from \n"
                "  - config path: '%s'\n"
                "  - config: '%s'",
                from_yaml,
                configuration,
            )

            return cast(
                AnnotationHandlerConfig,
                ConfigBuilder(
                    class_type=AnnotationHandlerConfig,
                    yaml_config_path=from_yaml,
                    string_replacement_map=string_replacement_map,
                ).configuration,
            )

        except ValueError as ve:
            logger.error(ve)
            raise ve

    @property
    def num_classes(self) -> int:
        """

        Returns:
            Number of classes the AnnotationHandler considers
        """
        return self.mapper.num_classes

    def parse_annotations_from_xml(self) -> List[BaseAnnotation]:
        """
        Creates a list of annotations using the configuration
        given via pascal_voc_input_data

        Returns:
            The created list
        """

        pascal_voc_parser = PascalVOCAnnotationParser(
            mapper=self.mapper,
            pascal_voc_input_data=self.configuration.pascal_voc_input_data,
        )

        annotations: List[BaseAnnotation] = pascal_voc_parser.parse()

        return annotations

    def parse_annotations_from_coco(self) -> List[BaseAnnotation]:
        """

        Returns: List of BaseAnnotations read from a coco formatted file (json)

        """

        coco_parser = COCOAnnotationParser(
            mapper=self.mapper, coco_input_data=self.configuration.coco_input_data
        )

        annotations: List[BaseAnnotation] = coco_parser.parse()

        return annotations

    def parse_annotations_from_label_studio(self) -> List[BaseAnnotation]:
        """

        Returns: List of BaseAnnotations read from a label studio formatted file (json)

        """

        label_studio_parser = LabelStudioAnnotationParser(
            mapper=self.mapper, label_studio_input_data=self.configuration.label_studio_input_data
        )

        annotations: List[BaseAnnotation] = label_studio_parser.parse()

        return annotations

    def parse_annotations_from_cvat(self) -> List[BaseAnnotation]:
        """

        Returns: List of BaseAnnotations read from a cvat formatted file (xml)

        """

        cvat_parser = CVATAnnotationParser(
            mapper=self.mapper, cvat_input_data=self.configuration.cvat_input_data
        )

        annotations: List[BaseAnnotation] = cvat_parser.parse()

        return annotations

    def parse_meta_info_from_cvat(self) -> List[ET_xml.Element]:
        """
        Returns:
            List of XMLElement objects that build the meta information in a CVAT formatted file
        """

        cvat_parser = CVATAnnotationParser(
            mapper=self.mapper, cvat_input_data=self.configuration.cvat_input_data
        )
        meta_info: List[ET_xml.Element] = cvat_parser.parse_cvat_meta_info()

        return meta_info

    def parse_annotations_from_csv(
        self,
        csv_file_path: str,
    ) -> List[BaseAnnotation]:
        """
        Create a dict of annotation information to a corresponding csv file
        The keys are the names of the classes_name_dict given by the classes_name_dict.
        Additionally there ist one key 'list' which stores a list of objects
        of the "BaseAnnotation" class

        Args:
            csv_file_path: path to csv file

        Returns:
            A List of BaseAnnotations which are loaded from the given csv

        """

        csv_parser = CSVAnnotationParser(
            mapper=self.mapper,
            csv_file_path=csv_file_path,
            pascal_voc_input_data=self.configuration.pascal_voc_input_data,
            coco_input_data=self.configuration.coco_input_data,
            cvat_input_data=self.configuration.cvat_input_data,
        )

        annotations: List[BaseAnnotation] = csv_parser.parse()

        return annotations

    def parse_annotations_from_mot(self) -> List[BaseAnnotation]:
        """

        Returns: List of BaseAnnotations read from a mot formatted file

        """

        mot_parser = MOTAnnotationParser(
            mapper=self.mapper, mot_input_data=self.configuration.mot_input_data
        )

        return mot_parser.parse()

    def generate_darknet_train_set(self, annotations: List[BaseAnnotation]) -> None:
        """
        Generate the dataset for the Darknet framework. It consists of two sets of images
        alongside with text files containing the annotations, one .txt per image. Two files,
        train.txt and test.txt, contain the list of image files.

        Args:
            annotations: List of BaseAnnotations from which training data is derived

        Returns:
            None
        """

        if self.configuration.write_output is None or (
            self.configuration.write_output is not None
            and self.configuration.write_output.darknet_train_set is None
        ):
            raise ValueError(
                "The write_output config is None! In order to be able to generate a darknet "
                "training set the write_output and write_output.darknet_train_set have to be "
                "provided!"
            )

        darknet_annotation_writer = DarknetAnnotationWriter(
            darknet_train_set_config=self.configuration.write_output.darknet_train_set,
            split_size=self.configuration.write_output.split_size,
        )

        _ = darknet_annotation_writer.write(annotations=annotations)

    def generate_csv(
        self,
        annotations: List[BaseAnnotation],
        output_string_format: str = CSVOutputStringFormats.BASE,
    ) -> Optional[str]:
        """
        Generate a csv file based on the given AnnotationHandler config.
        The generation is based on given directories for parsing image and annotation paths.
        Currently only .xml annotation files in PASCAL-VOC format are supported.

        Args:
            annotations: List of BaseAnnotations which are about to be transformed to csv format
            output_string_format: Format of generated csv string (one of CSVOutputStringFormats)

        Returns:
            Optional, path to the generated csv
        """

        assert self.configuration.write_output is not None
        assert self.configuration.write_output.csv_annotation is not None

        csv_annotation_writer = CSVAnnotationWriter(
            write_output_config=self.configuration.write_output,
            output_string_format=output_string_format,
        )

        output_file_path: Optional[str] = csv_annotation_writer.write(annotations=annotations)

        return output_file_path

    def parse_training_annotations(self) -> List[BaseAnnotation]:
        """
        Parse annotations from all different types of annotation formats that are provided
        by the CVAT export/import functionality. By setting the merge_content parameter to True,
        all annotations are merged on the basis of the image_path
        (which should be a unique identifier).

        Returns: a List of BaseAnnotations
        """

        annotations: List[BaseAnnotation] = []

        annotations.extend(self.parse_annotations_from_xml())
        annotations.extend(self.parse_annotations_from_coco())
        annotations.extend(self.parse_annotations_from_cvat())
        annotations.extend(self.parse_annotations_from_mot())
        annotations.extend(self.parse_annotations_from_label_studio())

        return annotations

    def parse_inference_annotations(self) -> List[BaseAnnotation]:
        """
        Parse annotations from all different types of annotation formats that are provided
        by the CVAT export/import functionality. By setting the merge_content parameter to True,
        all annotations are merged on the basis of the image_path
        (which should be a unique identifier).

        Returns: a List of BaseAnnotations
        """

        return self.parse_training_annotations()

    @staticmethod
    def reduce_annotations(
        annotations: List[BaseAnnotation], mapper: AnnotationClassMapper
    ) -> List[BaseAnnotation]:
        mapped_annotations: List[BaseAnnotation] = []
        for annotation in annotations:
            mapped_classifications: List[Classification] = []
            mapped_bounding_boxes: List[BoundingBox] = []
            mapped_segmentations: List[Segmentation] = []

            for classification in annotation.classifications:
                class_identifiers = mapper.map_model_class_id_to_output_class_identifier(
                    class_id=classification.class_id
                )

                for class_identifier in class_identifiers:
                    mapped_classifications.append(
                        classification.copy_classification(
                            class_identifier=class_identifier,
                        )
                    )

            for bounding_box in annotation.bounding_boxes:
                class_identifiers = mapper.map_model_class_id_to_output_class_identifier(
                    class_id=bounding_box.class_id
                )

                for class_identifier in class_identifiers:
                    mapped_bounding_boxes.append(
                        bounding_box.copy_bounding_box(
                            class_identifier=class_identifier,
                        )
                    )

            for segmentation in annotation.segmentations:
                class_identifiers = mapper.map_model_class_id_to_output_class_identifier(
                    class_id=segmentation.class_id
                )

                for class_identifier in class_identifiers:
                    mapped_segmentations.append(
                        segmentation.copy_segmentation(
                            class_identifier=class_identifier,
                        )
                    )

            mapped_annotations.append(
                annotation.copy_annotation(
                    classifications=mapped_classifications,
                    bounding_boxes=mapped_bounding_boxes,
                    segmentations=mapped_segmentations,
                )
            )

        return mapped_annotations



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing PascalVOC formatted annotations"""
import logging
import os
from typing import List, Tuple
from xml.etree.ElementTree import ParseError

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerPASCALVOCInputDataConfig,
)
from mlcvzoo_base.configuration.structs import AnnotationFileFormats, ImageFileFormats
from mlcvzoo_base.data_preparation.annotation_builder.pascal_voc_annotation_builder import (
    PascalVOCAnnotationBuilder,
)
from mlcvzoo_base.utils.file_utils import get_file_list

logger = logging.getLogger(__name__)


class PascalVOCAnnotationParser(AnnotationParser):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        pascal_voc_input_data: List[AnnotationHandlerPASCALVOCInputDataConfig],
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.pascal_voc_input_data = pascal_voc_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = list()
        # Fill dictionary for implicit path replacement
        for dataset_count, input_data_config in enumerate(self.pascal_voc_input_data):
            replacement_string = AnnotationParser.csv_directory_replacement_string.format(
                dataset_count
            )

            image_path_list, annotation_path_list = self.__determine_data_paths(
                input_data_config=input_data_config
            )

            if input_data_config.ignore_missing_images:
                image_path_list = len(annotation_path_list) * [""]
            else:
                (image_path_list, annotation_path_list,) = self.__ensure_annotations_match_images(
                    image_path_list=image_path_list,
                    annotation_path_list=annotation_path_list,
                )

            for image_path, annotation_path in zip(image_path_list, annotation_path_list):
                if (
                    not input_data_config.ignore_missing_images
                    and os.path.realpath(input_data_config.input_image_dir)
                    in os.path.realpath(image_path)
                    and os.path.realpath(input_data_config.input_xml_dir)
                    in os.path.realpath(annotation_path)
                ) or (
                    input_data_config.ignore_missing_images
                    and os.path.realpath(input_data_config.input_xml_dir)
                    in os.path.realpath(annotation_path)
                ):
                    try:
                        pascal_voc_builder = PascalVOCAnnotationBuilder(
                            mapper=self.mapper,
                            use_difficult=input_data_config.use_difficult,
                        )

                        annotation: BaseAnnotation = pascal_voc_builder.build(
                            image_path=image_path,
                            annotation_path=annotation_path,
                            image_dir=input_data_config.input_image_dir,
                            annotation_dir=input_data_config.input_xml_dir,
                            replacement_string=replacement_string,
                        )

                        annotations.append(annotation)

                    except (ForbiddenClassError, ParseError) as error:
                        logger.warning(
                            "%s: %s, annotation '%s' will be skipped"
                            % (type(error), str(error), annotation_path),
                        )
                        continue

        return annotations

    @staticmethod
    def __determine_data_paths(
        input_data_config: AnnotationHandlerPASCALVOCInputDataConfig,
    ) -> Tuple[List[str], List[str]]:
        """
        Determine all image and annotations paths from the given input-data-config

        Args:
            input_data_config: The AnnotationHandlerPASCALVOCInputDataConfig that defines
                               the input-data

        Returns:
            Tuple of image paths and annotation paths
        """

        image_paths = []
        xml_paths = []

        # CHECK SUBDIR CONFIG
        if len(input_data_config.input_sub_dirs) > 0:
            for sub_input_dir in input_data_config.input_sub_dirs:
                # READ IMAGES
                input_image_dir = os.path.join(input_data_config.input_image_dir, sub_input_dir)

                image_paths.extend(
                    get_file_list(
                        input_dir=input_image_dir,
                        search_subfolders=True,
                        file_extension=f"{input_data_config.image_format}",
                    )
                )

                # READ ANNOTATIONS
                input_xml_dir = os.path.join(input_data_config.input_xml_dir, sub_input_dir)

                xml_paths.extend(
                    get_file_list(
                        input_dir=input_xml_dir,
                        search_subfolders=True,
                        # exclude_pattern="predicted",
                        file_extension=AnnotationFileFormats.XML,
                    )
                )
        else:
            # READ IMAGES
            image_paths.extend(
                get_file_list(
                    input_dir=input_data_config.input_image_dir,
                    search_subfolders=True,
                    file_extension=input_data_config.image_format,
                )
            )

            # READ ANNOTATIONS
            xml_paths.extend(
                get_file_list(
                    input_dir=input_data_config.input_xml_dir,
                    search_subfolders=True,
                    # exclude_pattern="predicted",
                    file_extension=AnnotationFileFormats.XML,
                )
            )

        image_paths.sort()
        xml_paths.sort()

        return image_paths, xml_paths

    @staticmethod
    def __ensure_annotations_match_images(
        image_path_list: List[str], annotation_path_list: List[str]
    ) -> Tuple[List[str], List[str]]:
        """

        Args:
            image_path_list:
            annotation_path_list:

        Returns:

        """
        # remove duplicate
        image_path_list = list(dict.fromkeys(image_path_list))
        annotation_path_list = list(dict.fromkeys(annotation_path_list))

        # TODO: allow more image formats
        cleaned_img = [
            os.path.basename(image_path)
            .replace(ImageFileFormats.JPEG, "")
            .replace(ImageFileFormats.PNG, "")
            for image_path in image_path_list
        ]
        cleaned_xml = [
            os.path.basename(annotation_path)
            .replace(AnnotationFileFormats.XML, "")
            .replace("_predicted", "")
            for annotation_path in annotation_path_list
        ]

        in_both_lists = set(cleaned_xml) & set(cleaned_img)

        img_indices = [idx for (idx, x) in enumerate(cleaned_img) if x in in_both_lists]
        xml_indices = [idx for (idx, x) in enumerate(cleaned_xml) if x in in_both_lists]

        cleaned_image_path_list = [
            image_path
            for (img_index, image_path) in enumerate(image_path_list)
            if img_index in img_indices
        ]
        cleaned_annotation_path_list = [
            annotation_path
            for (annotation_index, annotation_path) in enumerate(annotation_path_list)
            if annotation_index in xml_indices
        ]

        if len(image_path_list) > len(cleaned_image_path_list):
            logger.debug(
                "Did not find an annotation-paths for image-paths: \n\t- %s\n",
                "\n\t- ".join(
                    [
                        image_path_list[index]
                        for index in range(0, len(image_path_list))
                        if index not in img_indices
                    ]
                ),
            )

        if len(annotation_path_list) > len(cleaned_annotation_path_list):
            logger.debug(
                "Did not find an image-paths for annotation-paths: \n\t- %s\n",
                "\n\t- ".join(
                    [
                        annotation_path_list[index]
                        for index in range(0, len(annotation_path_list))
                        if index not in xml_indices
                    ]
                ),
            )

        return cleaned_image_path_list, cleaned_annotation_path_list



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for parsing information from yaml in python accessible attributes for using mlflow.
"""
from typing import Dict, List, Optional

import related
from attr import define
from config_builder import BaseConfigClass


@define
class MLFlowPostgresSQLConfig(BaseConfigClass):
    """Class for parsing information about a SQL sever used for storing experiment information"""

    database_user: str = related.StringField()
    database_pw: str = related.StringField()
    database_port: str = related.StringField(default=5432)
    database_name: str = related.StringField(default="mlflowdb")


@define
class MLFlowFileConfig(BaseConfigClass):
    """Class for parsing information about a directory used for storing experiment information"""

    logging_dir: str = related.StringField()


@define
class MLFlowConfig(BaseConfigClass):
    """
    Class for parsing general information about path handling and also further
    configuration information in respective hierarchy
    """

    artifact_location: str = related.StringField()

    mlflow_postgressql_config: Optional[MLFlowPostgresSQLConfig] = related.ChildField(
        cls=MLFlowPostgresSQLConfig, required=False, default=None
    )

    mlflow_file_config: Optional[MLFlowFileConfig] = related.ChildField(
        cls=MLFlowFileConfig, required=False, default=None
    )

    @property
    def _mutual_attributes(self) -> List[str]:
        return [
            "mlflow_postgressql_config",
            "mlflow_file_config",
        ]



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for configuring machine learning models"""

from abc import ABC
from typing import Optional

import related
from attr import define
from config_builder import BaseConfigClass


@define
class ModelConfiguration(BaseConfigClass, ABC):
    """
    A model configuration.
    Typically, subclasses of a configuration parser implement
    this class to provide a mechanism to feed parameters into models or training.
    """

    unique_name: str = related.StringField()


@define
class InferenceConfig(BaseConfigClass):
    """
    Overall inference config that defines parameters that
    are valid for every model that is performing an inference.
    """

    config_path: Optional[str] = related.ChildField(cls=str)
    checkpoint_path: str = related.StringField()
    score_threshold: float = related.FloatField()



# --------------------------------------------------------
# Fast R-CNN
# Copyright (c) 2015 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Sergey Karayev
# --------------------------------------------------------

"""
This module is used to source out a metric computing utility, which
is taken from https://github.com/rbgirshick/py-faster-rcnn
"""

import numpy as np


def voc_ap(rec, prec, use_07_metric=False):  # type: ignore
    """ap = voc_ap(rec, prec, [use_07_metric])
    Compute VOC AP given precision and recall.
    If use_07_metric is true, uses the
    VOC 07 11 point method (default:False).

    Code originally from
    https://github.com/rbgirshick/py-faster-rcnn/
    blob/master/lib/datasets/voc_eval.py#L31-L61
    """
    if use_07_metric:
        # 11 point metric
        ap = 0.0
        for t in np.arange(0.0, 1.1, 0.1):
            if np.sum(rec >= t) == 0:
                p = 0
            else:
                p = np.max(prec[rec >= t])
            ap = ap + p / 11.0
    else:
        # correct AP calculation
        # first append sentinel values at the end
        mrec = np.concatenate(([0.0], rec, [1.0]))
        mpre = np.concatenate(([0.0], prec, [0.0]))

        # compute the precision envelope
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # to calculate area under PR curve, look for points
        # where X axis (recall) changes value
        i = np.where(mrec[1:] != mrec[:-1])[0]

        # and sum (\Delta recall) * prec
        ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for different utility operations regarding xml documents and objects"""
import logging
import os
import xml.etree.ElementTree as ET_xml

from mlcvzoo_base.utils import ensure_dir

logger = logging.getLogger(__name__)


def create_xml_root(file_path: str, width: int, height: int, depth: str = "3") -> ET_xml.Element:
    """
    Creates the root-xml-structure, which is used to generate
    annotation files. These annotation contain information
    about the bounding boxes specified for a training image.

    The format of the xml-files is the same as the one used
    in PASCAL VOC: http://host.robots.ox.ac.uk/pascal/VOC/

    Args:
        file_path: String, complete path of the image, including the image name
        width: int, width of the image
        height: int, height of the image
        depth: String, channel of the images, 3 stands for color image

    Returns: the root object

    """

    root = ET_xml.Element("annotation")
    ET_xml.SubElement(root, "filename").text = os.path.basename(file_path)
    ET_xml.SubElement(root, "folder").text = os.path.basename(os.path.dirname(file_path))
    ET_xml.SubElement(root, "path").text = file_path

    source = ET_xml.SubElement(root, "source")
    ET_xml.SubElement(source, "database").text = "Unknown"

    # images size
    size = ET_xml.SubElement(root, "size")

    ET_xml.SubElement(size, "width").text = str(width)
    ET_xml.SubElement(size, "height").text = str(height)

    ET_xml.SubElement(size, "depth").text = depth

    ET_xml.SubElement(root, "segmented").text = "0"

    return root


def create_bbox_xml_entry(
    root: ET_xml.Element, name: str, xmin: int, ymin: int, xmax: int, ymax: int
) -> ET_xml.Element:
    """
    Creates an element in a xml-structure, containing the information
    about a bounding box.

    The format of the xml-files is the same as the one used
    in PASCAL VOC: http://host.robots.ox.ac.uk/pascal/VOC/

    xmin, ymin, xmax, ymax coordinates of the bounding box

    Args:
        root: XML Element, the root object, where the xml-structure should be put in
        name: String, class name of the bounding box
        xmin: int, xmin coordinate of bounding box
        ymin: int, ymin coordinate of bounding box
        xmax: int, xmax coordinate of bounding box
        ymax: int, ymax coordinate of bounding box

    Returns: XML Element, the new root object

    """

    object_entry = ET_xml.SubElement(root, "object")

    # objects
    ET_xml.SubElement(object_entry, "name").text = name
    ET_xml.SubElement(object_entry, "pose").text = "Unspecified"
    ET_xml.SubElement(object_entry, "truncated").text = "0"
    ET_xml.SubElement(object_entry, "difficult").text = "0"

    bbox_entry = ET_xml.SubElement(object_entry, "bndbox")

    # bounding boxes
    ET_xml.SubElement(bbox_entry, "xmin").text = str(int(xmin))
    ET_xml.SubElement(bbox_entry, "ymin").text = str(int(ymin))
    ET_xml.SubElement(bbox_entry, "xmax").text = str(int(xmax))
    ET_xml.SubElement(bbox_entry, "ymax").text = str(int(ymax))

    return root


def xml_tree_to_file(xml_file_path: str, xml_tree: ET_xml.ElementTree) -> None:
    """
    Writes XML object to given file

    Args:
        xml_file_path: String, file path
        xml_tree: XML ElementTree

    Returns: None

    """
    logger.info("Write xml-tree to file: %s", xml_file_path)

    ensure_dir(file_path=xml_file_path, verbose=True)

    tree_string = ET_xml.tostring(element=xml_tree.getroot()).decode("utf-8")

    with open(file=xml_file_path, mode="w", encoding="utf8") as xml_file:
        xml_file.write(tree_string)



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing COCO formatted annotations"""
import json
import logging
import os
from typing import List

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerSingleFileInputDataConfig,
)
from mlcvzoo_base.data_preparation.annotation_builder.coco_annotation_builder import (
    COCOAnnotationBuilder,
)

logger = logging.getLogger(__name__)


class COCOAnnotationParser(AnnotationParser):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        coco_input_data: List[AnnotationHandlerSingleFileInputDataConfig],
    ) -> None:
        AnnotationParser.__init__(self, mapper=mapper)

        self.coco_input_data = coco_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = list()

        for dataset_count, input_data in enumerate(self.coco_input_data):
            with open(file=input_data.input_path, mode="r", encoding="'utf-8") as annotation_file:
                annotation_dict = json.load(fp=annotation_file)

            coco_images = annotation_dict["images"]

            for coco_image in coco_images:
                image_path = os.path.join(input_data.input_root_dir, coco_image["file_name"])
                image_width = coco_image["width"]
                image_height = coco_image["height"]

                replacement_string = AnnotationParser.csv_directory_replacement_string.format(
                    dataset_count
                )

                coco_builder = COCOAnnotationBuilder(
                    image_shape=(image_height, image_width),
                    image_id=coco_image["id"],
                    coco_annotations=annotation_dict["annotations"],
                    coco_categories=annotation_dict["categories"],
                    mapper=self.mapper,
                    use_difficult=input_data.use_difficult,
                )

                try:
                    annotation = coco_builder.build(
                        image_path=image_path,
                        annotation_path=input_data.input_path,
                        image_dir=input_data.input_root_dir,
                        annotation_dir=os.path.dirname(input_data.input_path),
                        replacement_string=replacement_string,
                    )

                    annotations.append(annotation)

                except (ValueError, ForbiddenClassError) as error:
                    logger.warning("%s, annotation will be skipped" % str(error))
                    continue

        return annotations



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for enumerating options of annotation and file formats, as well as data subsets and
respective phases. Furthermore mlflow configuration options can be found here.
"""
from enum import Enum
from typing import Any, List


class BaseType:
    """Class for accessing internal attributes"""

    @staticmethod
    def get_values_as_list(class_type: Any) -> List[str]:
        list_types = []

        for attribute, value in class_type.__dict__.items():
            if (
                type(attribute) is str
                and not (attribute.startswith("__") and attribute.endswith("__"))
                and attribute != "get_values"
            ):
                list_types.append(value)
        return list_types

    @staticmethod
    def get_values_as_string(class_type: Any) -> str:
        types = ""

        for attribute, value in class_type.__dict__.items():
            if (
                type(attribute) is str
                and not (attribute.startswith("__") and attribute.endswith("__"))
                and attribute != "get_values"
            ):
                types += str(value) + ", "

        types = types[:-2]

        return types


class ObjectDetectionBBoxFormats(BaseType):
    """Class for enumerating accepted bounding box formats"""

    XYXY = "XYXY"  # => also known under xmin, ymin, xmax, ymax
    XYWH = "XYWH"


class AnnotationFormats(BaseType):
    """Class for enumerating accepted annotation formats"""

    BASE = "BASE"
    VOC_XML = "VOC_XML"
    CSV = "CSV"
    COCO = "COCO"
    CVAT_FOR_IMAGES = "CVAT_FOR_IMAGES"


class ImageFileFormats(BaseType):
    """Class for enumerating accepted image file formats"""

    JPEG = ".jpg"
    PNG = ".png"


class AnnotationFileFormats(BaseType):
    """Class for enumerating accepted annotation file formats"""

    XML = ".xml"


class ImageTypes(BaseType):
    """Class for enumerating accepted image object formats"""

    CV2 = "cv2"
    PIL = "pil"


class FileNamePlaceholders(BaseType):
    """Class for enumerating placeholders in filenames"""

    TIMESTAMP = "TIMESTAMP"
    IMAGE_NAME = "IMAGE_NAME"


class CSVFileNameIdentifier(BaseType):
    """Class for enumerating CSV annotation file identifiers regarding training phase"""

    TRAINING = "train"
    EVALUATION = "eval"
    VALIDATION = "validation"
    CROSS_VAL_SPLIT = "split"


class MMDetectionModelsTypes(BaseType):
    """Class for enumerating object detection model types used from mmdetection package"""

    HTC = "htc"


class MMOCRModelsTypes(BaseType):
    """Class for enumerating OCR model types used from mmocr package"""

    DBNET = "dbnet"


class PipelineStepTypes(BaseType):
    """Class for enumerating phases passed through a training process"""

    DATA_GENERATION = "DATA_GENERATION"
    TRAIN = "TRAIN"
    EVAL = "EVAL"
    CROSS_EVAL = "CROSS_EVAL"
    INFERENCE = "INFERENCE"


class MLFlowExperimentTypes(BaseType):
    """Class for enumerating types of mlflow experiments"""

    DATA_GENERATION = "DATA_GENERATION"
    TRAIN = "TRAIN"
    EVAL = "EVAL"
    TIMING = "TIMING"


class MLFlowExperimentConfig:
    """Class for configuring mlflow experiments"""

    EXPERIMENT_DICT = d = {
        index: i
        for index, i in enumerate(MLFlowExperimentTypes.get_values_as_list(MLFlowExperimentTypes))
    }


class MLFLowTrackingUriTypes(BaseType):
    """Class for enumerating supported mlflow tracking URI types"""

    FILE = "file"
    POSTGRES = "postgresql"


class OpenCVImageFormats(Enum):
    BMP = ".bmp"
    PBM = ".pbm"
    PGM = ".pgm"
    PPM = ".ppm"
    JPEG = ".jpeg"
    JPG = ".jpg"
    JPE = ".jpe"
    TIFF = ".tiff"
    TIF = ".tif"
    PNG = ".png"


class DeviceQueryTypes(Enum):
    NVIDIA_SMI = "nvidia_smi"



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

import logging
import os

import mlflow
from config_builder import BaseConfigClass

from mlcvzoo_base.configuration.utils import write_config_to_yaml
from mlcvzoo_base.utils.versioning_utils import get_git_info, write_pip_info_to_file

logger = logging.getLogger(__name__)


def mlflow_log_config_to_yaml(config: BaseConfigClass, output_yaml_config_path: str) -> None:
    """
    Save a given configuration to a yaml-file and log it to mlflow if specified

    Args:
        config:
        output_yaml_config_path:

    Returns:

    """

    write_config_to_yaml(config=config, output_yaml_config_path=output_yaml_config_path)

    if os.path.isfile(output_yaml_config_path):
        mlflow.log_artifact(
            local_path=output_yaml_config_path,
        )


def mlflow_log_pip_package_versions(output_requirements_path: str) -> None:
    """

    Args:
        output_requirements_path:

    Returns:

    """

    write_pip_info_to_file(output_requirements_path=output_requirements_path)

    if os.path.isfile(output_requirements_path):
        mlflow.log_artifact(
            local_path=output_requirements_path,
        )


def mlflow_log_git_info() -> None:
    """
    Log all git related information about the current project with mlflow

    Returns:

    """

    repo, sha, branch = get_git_info()

    mlflow.log_param(key="git_repo", value=str(repo))
    mlflow.log_param(key="git_sha", value=sha)
    mlflow.log_param(key="git_branch", value=str(branch))



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for custom exceptions"""


class ForbiddenClassError(Exception):
    """
    Error which indicates that an invalid class-name has been tried to be mapped
    """

    def __init__(self, invalid_class: str, postfix: str = ""):
        self.message: str = f"Parsed invalid class: '{invalid_class}' {postfix}"
        Exception.__init__(self, self.message)


class ClassMappingNotFoundError(Exception):
    """
    Error which indicates that a class ID / class name could not be mapped to
    a ClassIdentifier. This is relevant when parsing annotation information from
    annotation files like CVAT xml or COCO json or PASCAL VOC xml, or when the
    prediction output of a model is decoded.
    """



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for parsing CSV formatted annotations"""
import logging
import os
from dataclasses import field
from typing import List, Optional, Tuple

from tqdm import tqdm

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.annotation_parser import AnnotationParser
from mlcvzoo_base.api.exceptions import ForbiddenClassError
from mlcvzoo_base.configuration.annotation_handler_config import (
    AnnotationHandlerPASCALVOCInputDataConfig,
    AnnotationHandlerSingleFileInputDataConfig,
)
from mlcvzoo_base.data_preparation.annotation_builder.csv_annotation_builder import (
    CSVAnnotationBuilder,
)

logger = logging.getLogger(__name__)


class CSVAnnotationParser(AnnotationParser):
    """
    Super class for defining the methods that are needed to parse a list of
    instances that are of the type BaseAnnotation.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child AnnotationHandler class
    """

    def __init__(
        self,
        mapper: AnnotationClassMapper,
        csv_file_path: str,
        pascal_voc_input_data: List[AnnotationHandlerPASCALVOCInputDataConfig] = field(
            default_factory=lambda: []
        ),
        coco_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = field(
            default_factory=lambda: []
        ),
        cvat_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = field(
            default_factory=lambda: []
        ),
    ):
        AnnotationParser.__init__(self, mapper=mapper)

        self.csv_file_path = csv_file_path
        self.pascal_voc_input_data = pascal_voc_input_data
        self.coco_input_data = coco_input_data
        self.cvat_input_data = cvat_input_data

    def parse(self) -> List[BaseAnnotation]:
        annotations: List[BaseAnnotation] = list()

        if os.path.isfile(self.csv_file_path):
            with open(file=self.csv_file_path, mode="r", encoding="'utf-8") as annotations_file:
                csv_lines = annotations_file.readlines()

                for index in tqdm(range(len(csv_lines))):
                    csv_line = csv_lines[index].strip()

                    pascal_voc_input_data_index: Optional[int] = None
                    coco_input_data_index: Optional[int] = None
                    cvat_input_data_index: Optional[int] = None

                    pascal_voc_replacement_string: Optional[str] = None
                    coco_replacement_string: Optional[str] = None
                    cvat_replacement_string: Optional[str] = None

                    if len(self.pascal_voc_input_data) > 0:
                        (
                            pascal_voc_input_data_index,
                            pascal_voc_replacement_string,
                        ) = CSVAnnotationParser.__get_input_data_index(
                            dataset_len=len(self.pascal_voc_input_data),
                            csv_line=csv_line,
                        )

                    if len(self.coco_input_data) > 0:
                        (
                            coco_input_data_index,
                            coco_replacement_string,
                        ) = CSVAnnotationParser.__get_input_data_index(
                            dataset_len=len(self.coco_input_data),
                            csv_line=csv_line,
                        )

                    if len(self.cvat_input_data) > 0:
                        (
                            cvat_input_data_index,
                            cvat_replacement_string,
                        ) = CSVAnnotationParser.__get_input_data_index(
                            dataset_len=len(self.cvat_input_data),
                            csv_line=csv_line,
                        )

                    image_dir: Optional[str] = None
                    annotation_dir: Optional[str] = None
                    replacement_string: Optional[str] = None

                    if pascal_voc_input_data_index is not None:
                        image_dir = self.pascal_voc_input_data[
                            pascal_voc_input_data_index
                        ].input_image_dir

                        annotation_dir = self.pascal_voc_input_data[
                            pascal_voc_input_data_index
                        ].input_xml_dir

                        replacement_string = pascal_voc_replacement_string

                    elif coco_input_data_index is not None:
                        image_dir = self.coco_input_data[coco_input_data_index].input_root_dir
                        annotation_dir = self.coco_input_data[coco_input_data_index].input_path

                        replacement_string = coco_replacement_string

                    elif cvat_input_data_index is not None:
                        image_dir = self.cvat_input_data[cvat_input_data_index].input_root_dir
                        annotation_dir = self.cvat_input_data[cvat_input_data_index].input_path

                        replacement_string = cvat_replacement_string

                    if (
                        image_dir is not None
                        and annotation_dir is not None
                        and replacement_string is not None
                    ):
                        csv_builder = CSVAnnotationBuilder(csv_line=csv_line, mapper=self.mapper)

                        try:
                            annotation = csv_builder.build(
                                image_path="",
                                annotation_path="",
                                image_dir=image_dir,
                                annotation_dir=annotation_dir,
                                replacement_string=replacement_string,
                            )

                            annotations.append(annotation)

                        except (ValueError, ForbiddenClassError) as error:
                            logger.warning("%s, annotation will be skipped" % str(error))
                            continue
                    else:
                        logger.warning(
                            "Could not find valid replacement for image or annotation directory!\n"
                            "csv-line: %s \n"
                            "pascal-voc-len: %s \n"
                            "coco-len: %s \n"
                            "image_dir: %s \n"
                            "annotation_dir: %s \n"
                            % (
                                csv_line,
                                len(self.pascal_voc_input_data),
                                len(self.coco_input_data),
                                image_dir,
                                annotation_dir,
                            )
                        )
        else:
            error_string = f"ERROR: file '{self.csv_file_path}' does not exist!"

            logger.error(error_string)
            raise ValueError(error_string)

        return annotations

    @staticmethod
    def __get_input_data_index(
        dataset_len: int, csv_line: str
    ) -> Tuple[Optional[int], Optional[str]]:
        input_data_index: Optional[int] = None
        replacement_string: Optional[str] = None

        for i in range(0, dataset_len):
            replacement_string = AnnotationParser.csv_directory_replacement_string.format(i)

            logger.debug(
                "Check line: \n - replacement-string: %s \n - csv_line: %s ",
                replacement_string,
                csv_line,
            )

            if replacement_string is not None:
                image_path = csv_line.split(" ")[0]

                image_path_split = image_path.split("/")

                if image_path_split[0] == replacement_string:
                    input_data_index = i
                    break

        return input_data_index, replacement_string



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module that holds the class for storing and accessing annotation class mapping information."""

from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional

from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError, ForbiddenClassError
from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import (
    ReductionMappingConfig,
    ReductionMappingMappingConfig,
)

logger = logging.getLogger(__name__)


class DuplicateOutputClassError(Exception):
    """
    Error that indicates that an output class ID or output class name of
    an reduction mapping is not unique / has duplicates.
    """


class AnnotationClassMapper:
    """
    Class for handling the mapping of class ids to class names during the parsing of annotation
    files and the inference of models. It is defined by the class_mapping and reduction_mapping
    configurations that are handed over in the constructor, please see
    :class:`mlcvzoo_base.configuration.class_mapping_config.ClassMappingConfig` and
    :class:`mlcvzoo_base.configuration.reduction_mapping_config.ReductionMappingConfig`
    """

    UNKNOWN_CLASS_NAME: str = "unknown_{}"

    def __init__(
        self,
        class_mapping: ClassMappingConfig,
        reduction_mapping: Optional[ReductionMappingConfig] = None,
    ) -> None:
        self.class_mapping_config: ClassMappingConfig = class_mapping

        # Dictionary that defines how a class ID during the parsing of annotation files
        # are mapped to a model class names. Please see for examples
        #
        # EXAMPLE:
        # {
        #     0: "person",
        #     1: "truck",
        #     2: "car",
        #     3: "lion",
        # }
        self.__annotation_class_id_to_model_class_name_map: Dict[
            int, str
        ] = self.__create_annotation_class_id_to_model_class_name_map(
            class_mapping_config=class_mapping,
        )

        # Dictionary that defines how a class names during the parsing of annotation files
        # are mapped to a model class IDs.
        #
        # EXAMPLE:
        # {
        #     "person": 0,
        #     "truck": 1,
        #     "car": 2,
        #     "lion": 3,
        # }
        self.__annotation_class_name_to_model_class_id_map: Dict[
            str, int
        ] = self.__create_annotation_class_name_to_model_class_id_map(
            annotation_class_id_to_model_class_name_map=self.__annotation_class_id_to_model_class_name_map
        )

        # Dictionary for mapping annotation class names to model class names. This
        # can be used to fix typos or to aggregate classes to a single model class.
        # The
        #
        # EXAMPLE:
        # {
        #     "Person": "person",
        #     "LKW": "truck",
        #     "tuck": "truck",
        #     "PKW": "car",
        # }
        self.__annotation_class_name_to_model_class_name_map: Dict[
            str, str
        ] = self.__create_annotation_class_name_to_model_class_name_map(
            class_mapping_config=self.class_mapping_config
        )

        # Dictionary for mapping / reducing the model class IDs to an output class identifier.
        # The model class IDs 1 and 2 are reduced to class ID 1 and class name vehicle.
        # The model class ID is used to create multiple outputs.
        #
        # EXAMPLE:
        #
        # {
        #     1: [ClassIdentifier(class_id=1, class_name="vehicle")],
        #     2: [ClassIdentifier(class_id=1, class_name="vehicle")],
        #     3: [
        #         ClassIdentifier(class_id=15, class_name="animal"),
        #         ClassIdentifier(class_id=16, class_name="cat"),
        #     ],
        # }
        self.__model_class_id_to_class_identifier_map: Dict[
            int, List[ClassIdentifier]
        ] = self.__create_model_class_id_to_class_identifier_map(
            annotation_class_id_to_model_class_name_map=self.__annotation_class_id_to_model_class_name_map,
            annotation_class_name_to_model_class_id_map=self.__annotation_class_name_to_model_class_id_map,
            reduction_mapping_config=reduction_mapping,
        )

        # Dictionary for mapping / reducing the model class names to an output class identifier.
        # The model class names "banana" and "cherry" are reduced to class ID 10
        # and class name "fruit".
        #
        # EXAMPLE:
        #
        # {
        #     "banana": [ClassIdentifier(class_id=10, class_name="fruit")],
        #     "cherry": [ClassIdentifier(class_id=10, class_name="fruit")],
        #     "lion": [
        #         ClassIdentifier(class_id=15, class_name="animal"),
        #         ClassIdentifier(class_id=16, class_name="cat"),
        #     ],
        # }
        self.__model_class_name_to_class_identifier_map: Dict[
            str, List[ClassIdentifier]
        ] = self.__create_model_class_name_to_class_identifier_map(
            annotation_class_id_to_model_class_name_map=self.__annotation_class_id_to_model_class_name_map,
            model_class_id_to_class_identifier_map=self.__model_class_id_to_class_identifier_map,
        )

        logger.debug("AnnotationClassMapper initialized")

    @property
    def num_classes(self) -> int:
        """
        Returns the number of classes the AnnotationMapper 'knows'.

        Returns:
            The number of classes the AnnotationMapper 'knows'.
        """

        return self.class_mapping_config.get_number_classes()

    @property
    def annotation_class_id_to_model_class_name_map(self) -> Dict[int, str]:
        return self.__annotation_class_id_to_model_class_name_map

    def create_class_identifier_list(self) -> List[ClassIdentifier]:
        """
        Build a list of all (output) ClassIdentifiers that are defined for this
        AnnotationClassMapper instance. Note that one Class-ID/Class-Name can
        have multiple mappings to ClassIdentifiers and multiple Class-IDs/Class-Names
        can share the same ClassIdentifier.

        Example mapping:
        {
            1: [ClassIdentifier(class_id=1, class_name="vehicle")],
            2: [ClassIdentifier(class_id=1, class_name="vehicle")],
            3: [
                ClassIdentifier(class_id=15, class_name="animal"),
                ClassIdentifier(class_id=16, class_name="cat"),
            ],
        }

        Produces:
        [
            ClassIdentifier(class_id=1, class_name="vehicle"),
            ClassIdentifier(class_id=15, class_name="animal"),
            ClassIdentifier(class_id=16, class_name="cat")
        ]

        Returns:
            The created list of ClassIdentifiers
        """
        class_identifier_list: List[ClassIdentifier] = []

        for _class_identifier_list in self.__model_class_id_to_class_identifier_map.values():
            class_identifier_list.extend(_class_identifier_list)

        return class_identifier_list

    def get_model_class_names(self) -> List[str]:
        """
        Returns:
            All model class names as list.
        """

        class_ids: List[int] = list(self.__annotation_class_id_to_model_class_name_map.keys())

        # We have to provide the same class names we use later to
        # reverse translate to class ids. Python guarantees the order
        # of insertion, not alphabetical or other orders, so we have
        # to sort the keys
        class_ids.sort()

        class_names: List[str] = []
        for class_id in class_ids:
            class_names.append(self.__annotation_class_id_to_model_class_name_map[class_id])

        return class_names

    @staticmethod
    def get_coco_classes_id_dict(categories: List[Dict[str, Any]]) -> Dict[int, str]:
        """

        Args:
            categories: Dictionary of COCO formatted category (or class) ids and names

        Returns:
            Dictionary of class id values to class names
        """

        __classes_id_dict: Dict[int, str] = dict()

        for category in categories:
            class_id = category["id"]
            class_name = category["name"]

            __classes_id_dict[class_id] = class_name

        return __classes_id_dict

    def map_annotation_class_name_to_model_class_name(self, class_name: str) -> str:
        """
        Map the given class name to a model class name according to
        the defined mapping in the configuration.

        Args:
            class_name: The class name to map

        Returns:
            The mapped class name

        Raises:
            ForbiddenClassError when the class_name matches a class name in the
            ignore_class_names list of the configuration.

            ClassMappingNotFoundError when the class_name can not be mapped to a model class
            name.
        """

        if not self.is_valid_class(class_name=class_name):
            raise ForbiddenClassError(invalid_class=class_name)

        # Try to find a valid mapping for the given input-key
        if class_name in self.__annotation_class_name_to_model_class_name_map:
            return self.__annotation_class_name_to_model_class_name_map[class_name]

        raise ClassMappingNotFoundError(
            f"Could not find a valid mapping for class-name='{class_name}'. "
            f"Please add it in the mapping configuration or add it to the configuration list "
            f"'ignore_class_names' "
        )

    def map_annotation_class_name_to_model_class_id(self, class_name: str) -> int:
        """
        Map the given class name to a model class ID according to
        the defined mapping in the configuration.

        Args:
            class_name: The class name to map

        Returns:
            The mapped class ID

        Raises:
            ForbiddenClassError when the class_name matches a class name in the
            ignore_class_names list of the configuration.

            ClassMappingNotFoundError when the class_name can not be mapped to a model class
            name.
        """

        model_class_name = self.map_annotation_class_name_to_model_class_name(
            class_name=class_name
        )

        # Try to find a valid mapping for the given input-key
        if model_class_name in self.__annotation_class_name_to_model_class_id_map:
            return self.__annotation_class_name_to_model_class_id_map[model_class_name]

        raise ClassMappingNotFoundError(
            f"Could not find a valid mapping for class-name='{class_name}'"
        )

    def map_annotation_class_id_to_model_class_name(self, class_id: int) -> str:
        """
        Map the given class id to a model class name according to
        the defined mapping in the configuration.

        Args:
            class_id: The class id to map

        Returns:
            The mapped class name

        Raises:
            ClassMappingNotFoundError when the class_name can not be mapped to a model class
            name.
        """

        # Try to find a valid mapping for the given input-key
        if class_id in self.__annotation_class_id_to_model_class_name_map:
            return self.__annotation_class_id_to_model_class_name_map[class_id]

        raise ClassMappingNotFoundError(
            f"Could not find a valid mapping for class-id='{class_id}'"
        )

    def map_model_class_id_to_output_class_identifier(
        self, class_id: int
    ) -> List[ClassIdentifier]:
        """
        Map the given class id to a list of ClassIdentifiers according to
        the defined mapping in the configuration.

        Args:
            class_id: The class id to map

        Returns:
            The list of ClassIdentifiers

        Raises:
            ClassMappingNotFoundError when the class_name can not be mapped to a model class
            name.
        """

        # Try to find a valid mapping for the given input-key
        if class_id in self.__model_class_id_to_class_identifier_map:
            return self.__model_class_id_to_class_identifier_map[class_id]

        raise ClassMappingNotFoundError(
            f"Could not find a valid mapping for class-id='{class_id}'"
        )

    def map_model_class_name_to_output_class_identifier(
        self, class_name: str
    ) -> List[ClassIdentifier]:
        """
        Map the given class name to a list of ClassIdentifiers according to
        the defined mapping in the configuration.

        Args:
            class_name: The class name to map

        Returns:
            The list of ClassIdentifiers

        Raises:
            ClassMappingNotFoundError when the class_name can not be mapped to a model class
            name.
        """

        # Try to find a valid mapping for the given input-key
        if class_name in self.__model_class_name_to_class_identifier_map:
            return self.__model_class_name_to_class_identifier_map[class_name]

        raise ClassMappingNotFoundError(
            f"Could not find a valid mapping for class-name='{class_name}'"
        )

    def is_valid_class(self, class_name: str) -> bool:
        """
        Checks whether a class name should be ignored by the AnnotationClassMapper

        Args:
            class_name: String, a name of a class

        Returns:
            Bool, true if the name is not listed in the 'ignore_class_name' attribute
        """

        # A class is only valid, if is it not contained in the
        # ignore_class_names list.
        if class_name in self.class_mapping_config.ignore_class_names:
            return False
        else:
            return True

    @staticmethod
    def __create_annotation_class_name_to_model_class_name_map(
        class_mapping_config: ClassMappingConfig,
    ) -> Dict[str, str]:
        """
        Create a dictionary for mapping a class-name as string that is parsed from annotation data,
        to the respective model class-name as string.

        There are the following restrictions:

        The attribute class_mapping_config.mapping can only define mappings to
        class names are specified in the attribute class_mapping_config.model_classes.

        Args:
            class_mapping_config: configuration object that provides the relevant
                                  mapping information

        Returns:
            A dictionary representing the mapping
        """

        class_name_mapping: Dict[str, str] = {}

        model_class_names: List[str] = []

        for model_class in class_mapping_config.model_classes:
            model_class_names.append(model_class.class_name)
            class_name_mapping[model_class.class_name] = model_class.class_name

        for index, mapping in enumerate(class_mapping_config.mapping):
            if mapping.output_class_name in model_class_names:
                class_name_mapping[mapping.input_class_name] = mapping.output_class_name
            else:
                raise ValueError(
                    "Invalid mapping config entry: "
                    "mapping[%d].output_class_name='%s', "
                    "but has to be one of '%s'"
                    % (index, mapping.output_class_name, model_class_names)
                )

        return class_name_mapping

    @staticmethod
    def __create_annotation_class_id_to_model_class_name_map(
        class_mapping_config: ClassMappingConfig,
    ) -> Dict[int, str]:
        """
        Create a dictionary for mapping annotation class IDs to model class names.

        There are the following features and restrictions:

        In case the configuration attribute class_mapping_config.model_classes does not contain
        an entry for every class ID in the range of 0, ..., num_classes, a default mapping
        will be created the maps the respective class ID to the string "unknown_ID".

        The configuration attribute class_mapping_config.model_classes can only contain class IDs
        that are in the range of 0, ..., num_classes.

        Args:
            class_mapping_config: The configuration object that defines the mapping

        Returns:
            The created dictionary
        """

        invalid_class_ids: List[int] = []
        classes_id_dict: Dict[int, str] = {}

        for model_class in class_mapping_config.model_classes:
            if model_class.class_id >= class_mapping_config.get_number_classes():
                invalid_class_ids.append(model_class.class_id)
                continue

            if model_class.class_id not in classes_id_dict:
                classes_id_dict[model_class.class_id] = model_class.class_name
            else:
                raise ValueError(
                    "Duplicate class_id='%s' is not allowed in model_classes"
                    % model_class.class_id
                )

        # Generate unknown classes string. This ensures that one
        # is not forced to define all class-names for the maximum number of classes
        for class_id in range(0, class_mapping_config.get_number_classes()):
            if class_id not in classes_id_dict:
                classes_id_dict[class_id] = AnnotationClassMapper.UNKNOWN_CLASS_NAME.format(
                    class_id
                )

        if len(invalid_class_ids) > 0:
            raise ValueError(
                "Found class_ids='%s' that exceed the number_classes=%d"
                % (invalid_class_ids, class_mapping_config.get_number_classes())
            )

        return classes_id_dict

    @staticmethod
    def __create_annotation_class_name_to_model_class_id_map(
        annotation_class_id_to_model_class_name_map: Dict[int, str]
    ) -> Dict[str, int]:
        """
        Create a dictionary for mapping annotation class names to model class IDs.

        It is the reverse of an annotation_class_id_to_model_class_name_map.

        Args:
            annotation_class_id_to_model_class_name_map: A dictionary that defines a mapping from
                                                         annotation class IDs to model class names

        Returns:
            The reversed annotation_class_id_to_model_class_name_map
        """

        return dict(
            map(reversed, annotation_class_id_to_model_class_name_map.items())  # type: ignore
        )

    @staticmethod
    def __create_model_class_id_to_class_identifier_map(
        annotation_class_id_to_model_class_name_map: Dict[int, str],
        annotation_class_name_to_model_class_id_map: Dict[str, int],
        reduction_mapping_config: Optional[ReductionMappingConfig],
    ) -> Dict[int, List[ClassIdentifier]]:
        default_model_class_id_to_class_identifier_map: Dict[int, List[ClassIdentifier]] = {}
        model_class_id_to_class_identifier_map: Dict[int, List[ClassIdentifier]] = {}

        for (
            class_id,
            model_class_name,
        ) in annotation_class_id_to_model_class_name_map.items():
            default_model_class_id_to_class_identifier_map[class_id] = [
                ClassIdentifier(class_id=class_id, class_name=model_class_name)
            ]

        output_class_id_map: Dict[int, ReductionMappingMappingConfig] = {}
        output_class_name_map: Dict[str, ReductionMappingMappingConfig] = {}

        if reduction_mapping_config is not None:
            for index, mapping_entry in enumerate(reduction_mapping_config.mapping):
                if mapping_entry.output_class_id in output_class_id_map:
                    raise DuplicateOutputClassError(
                        "Duplicate definition for output-class-id='%s' "
                        "for mapping_entry at index='%s':\n"
                        " - already stored mapping-entry='%s'\n"
                        " - current mapping-entry='%s'"
                        % (
                            mapping_entry.output_class_id,
                            index,
                            output_class_id_map[mapping_entry.output_class_id],
                            mapping_entry,
                        ),
                    )

                if mapping_entry.output_class_name in output_class_name_map:
                    raise DuplicateOutputClassError(
                        "Duplicate definition for output-class-name='%s' "
                        "for mapping_entry at index='%s':\n"
                        " - already stored mapping-entry='%s'\n"
                        " - current mapping-entry='%s'"
                        % (
                            mapping_entry.output_class_name,
                            index,
                            output_class_name_map[mapping_entry.output_class_name],
                            mapping_entry,
                        ),
                    )

                output_class_id_map[mapping_entry.output_class_id] = mapping_entry
                output_class_name_map[mapping_entry.output_class_name] = mapping_entry

                if mapping_entry.model_class_ids is not None:
                    model_class_id_to_class_identifier_map.update(
                        AnnotationClassMapper.__update_model_class_id_to_class_identifier_map_from_model_class_ids(
                            model_class_id_to_class_identifier_map=model_class_id_to_class_identifier_map,
                            annotation_class_id_to_model_class_name_map=annotation_class_id_to_model_class_name_map,
                            mapping_entry=mapping_entry,
                            index=index,
                        )
                    )

                if mapping_entry.model_class_names is not None:
                    model_class_id_to_class_identifier_map.update(
                        AnnotationClassMapper.__update_model_class_id_to_class_identifier_map_from_model_class_names(
                            model_class_id_to_class_identifier_map=model_class_id_to_class_identifier_map,
                            annotation_class_name_to_model_class_id_map=annotation_class_name_to_model_class_id_map,
                            mapping_entry=mapping_entry,
                            index=index,
                        )
                    )

        default_model_class_id_to_class_identifier_map.update(
            model_class_id_to_class_identifier_map
        )

        return default_model_class_id_to_class_identifier_map

    @staticmethod
    def __update_model_class_id_to_class_identifier_map_from_model_class_ids(
        model_class_id_to_class_identifier_map: Dict[int, List[ClassIdentifier]],
        annotation_class_id_to_model_class_name_map: Dict[int, str],
        mapping_entry: ReductionMappingMappingConfig,
        index: int,
    ) -> Dict[int, List[ClassIdentifier]]:
        assert mapping_entry.model_class_ids is not None

        error_class_ids: List[int] = []
        # Any model-class-id defines a mapping from class-id to a ClassIdentifier
        for model_class_id in mapping_entry.model_class_ids:
            if model_class_id not in annotation_class_id_to_model_class_name_map:
                error_class_ids.append(model_class_id)
                continue

            class_identifier = ClassIdentifier(
                class_id=mapping_entry.output_class_id,
                class_name=mapping_entry.output_class_name,
            )

            if model_class_id not in model_class_id_to_class_identifier_map:
                model_class_id_to_class_identifier_map[model_class_id] = [class_identifier]
            else:
                model_class_id_to_class_identifier_map[model_class_id].append(class_identifier)

        if len(error_class_ids) > 0:
            raise ValueError(
                "Invalid reduction-mapping config entry:\n"
                " - mapping[%d].model_class_ids='%s'\n"
                " - it can only contain values of: '%s'\n"
                " - wrong model_class_ids: '%s'"
                % (
                    index,
                    mapping_entry.model_class_ids,
                    list(annotation_class_id_to_model_class_name_map.keys()),
                    error_class_ids,
                )
            )

        return model_class_id_to_class_identifier_map

    @staticmethod
    def __update_model_class_id_to_class_identifier_map_from_model_class_names(
        model_class_id_to_class_identifier_map: Dict[int, List[ClassIdentifier]],
        annotation_class_name_to_model_class_id_map: Dict[str, int],
        mapping_entry: ReductionMappingMappingConfig,
        index: int,
    ) -> Dict[int, List[ClassIdentifier]]:
        assert mapping_entry.model_class_names is not None

        error_class_names: List[str] = []
        for model_class_name in mapping_entry.model_class_names:
            if model_class_name in annotation_class_name_to_model_class_id_map:
                model_class_id = annotation_class_name_to_model_class_id_map[model_class_name]
            else:
                error_class_names.append(model_class_name)
                continue

            class_identifier = ClassIdentifier(
                class_id=mapping_entry.output_class_id,
                class_name=mapping_entry.output_class_name,
            )

            if model_class_id not in model_class_id_to_class_identifier_map:
                model_class_id_to_class_identifier_map[model_class_id] = [class_identifier]
            else:
                model_class_id_to_class_identifier_map[model_class_id].append(class_identifier)

        if len(error_class_names) > 0:
            raise ValueError(
                "Invalid reduction-mapping config entry:\n"
                " - mapping[%d].model_class_names='%s'\n"
                " - it can only contain values of: '%s'\n"
                " - wrong model_class_names: '%s'"
                % (
                    index,
                    mapping_entry.model_class_names,
                    list(annotation_class_name_to_model_class_id_map.keys()),
                    error_class_names,
                )
            )

        return model_class_id_to_class_identifier_map

    @staticmethod
    def __create_model_class_name_to_class_identifier_map(
        annotation_class_id_to_model_class_name_map: Dict[int, str],
        model_class_id_to_class_identifier_map: Dict[int, List[ClassIdentifier]],
    ) -> Dict[str, List[ClassIdentifier]]:
        return {
            annotation_class_id_to_model_class_name_map[model_class_id]: class_identifier
            for model_class_id, class_identifier in model_class_id_to_class_identifier_map.items()
        }



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for parsing information from yaml in python accessible attributes for the
AnnotationHandler class.
"""
import logging
import os
from dataclasses import field
from typing import Dict, List, Optional

import related
from attr import define
from config_builder import BaseConfigClass

from mlcvzoo_base.configuration.class_mapping_config import ClassMappingConfig
from mlcvzoo_base.configuration.reduction_mapping_config import ReductionMappingConfig
from mlcvzoo_base.data_preparation.structs import (
    CSVOutputStringFormats,
    MOTChallengeFormats,
)

logger = logging.getLogger(__name__)


@define
class AnnotationHandlerPASCALVOCInputDataConfig(BaseConfigClass):
    """Class for parsing information of PASCAL VOC input format"""

    __related_strict__ = True

    input_image_dir: str = related.StringField()
    input_xml_dir: str = related.StringField()
    input_sub_dirs: List[str] = related.SequenceField(str)

    image_format: str = related.StringField()
    use_difficult: bool = related.BooleanField(default=True)
    ignore_missing_images: bool = related.BooleanField(default=False)


@define
class AnnotationHandlerMOTInputDataConfig(BaseConfigClass):
    """Class for parsing MOT information of input data"""

    __related_strict__ = True

    # Path to the annotation file
    annotation_path: str = related.StringField()
    # Root directory which fits to the relative directory in the given annotation file
    image_dir: str = related.StringField()
    # considered image file format
    image_format: str = related.StringField(required=False, default=".jpg")
    # specifies which format of which challenge/year should be used
    mot_format: str = related.StringField(required=False, default=MOTChallengeFormats.MOT20.value)
    # specifies how to interpret confidence scores in annotations
    ground_truth: bool = related.BooleanField(required=False, default=True)
    # path to the files where the MOT labels are specified
    label_path: Optional[str] = related.ChildField(cls=str, required=False, default=None)

    def check_values(self) -> bool:
        return self.mot_format in [f.value for f in MOTChallengeFormats]


@define
class AnnotationHandlerLabelStudioInputDataConfig(BaseConfigClass):
    """Class for parsing information of input data"""

    input_image_dir: str = related.StringField()
    input_annotation_dir: str = related.StringField()

    image_format: str = related.StringField()
    # TODO: don't set defaults here?
    use_difficult: bool = related.BooleanField(default=True)
    ignore_missing_images: bool = related.BooleanField(default=False)


@define
class AnnotationHandlerSingleFileInputDataConfig(BaseConfigClass):
    """Class for parsing information of input data"""

    __related_strict__ = True

    # Path to the annotation file
    input_path: str = related.StringField()
    # Root directory which fits to the relative directory in the given annotation file
    # TODO: Optional?
    input_root_dir: str = related.StringField()

    # TODO: don't set defaults here?
    use_difficult: bool = related.BooleanField(default=True)


@define
class AnnotationHandlerWriteOutputCSVAnnotationConfig(BaseConfigClass):
    """Class for parsing information of output format when writing CSV annotations"""

    __related_strict__ = True

    csv_dir: str = related.StringField()
    csv_base_file_name: str = related.StringField(default="")

    csv_split_0_file_name: str = related.StringField(default="")
    csv_split_1_file_name: str = related.StringField(default="")

    # TODO: add hint for file-name replacement? Add specific config class?
    timestamp_format: str = related.StringField(default="%Y-%m-%d")
    include_surrounding_bboxes: bool = related.BooleanField(default=True)
    output_string_format: str = related.StringField(default=CSVOutputStringFormats.BASE)

    def check_values(self) -> bool:
        if not os.path.isdir(self.csv_dir):
            # Note: this is not necessarily required or
            #       already existing when config is read (see test)
            logger.info("Class='%s', specify correct output_dir'%s'", type(self), self.csv_dir)

        return self.output_string_format in CSVOutputStringFormats.get_values_as_list(
            class_type=CSVOutputStringFormats
        )


@define
class AnnotationHandlerWriteOutputDarknetAnnotationConfig(BaseConfigClass):
    """Class for parsing information of output format when writing Darknet annotations"""

    __related_strict__ = True

    train_data_set_dir: str = related.StringField()
    base_file_name: str = related.StringField(default="")
    write_as_symlink: bool = related.BooleanField(default=False)

    def get_train_file_path(self) -> str:
        return os.path.join(self.train_data_set_dir, f"{self.base_file_name}_train.txt")

    def get_test_file_path(self) -> str:
        return os.path.join(self.train_data_set_dir, f"{self.base_file_name}_test.txt")


@define
class AnnotationHandlerWriteOutputBoxSnippetsConfig(BaseConfigClass):
    """Class for parsing information of output directory when writing annotations"""

    __related_strict__ = True

    output_dir: str = related.StringField()


@define
class AnnotationHandlerWriteOutputConfig(BaseConfigClass):
    """Class for parsing information of output format when writing annotations"""

    __related_strict__ = True

    use_cross_val: bool = related.BooleanField()
    number_splits: int = related.IntegerField()
    split_size: int = related.FloatField()

    csv_annotation: Optional[AnnotationHandlerWriteOutputCSVAnnotationConfig] = related.ChildField(
        cls=AnnotationHandlerWriteOutputCSVAnnotationConfig,
        required=False,
        default=None,
    )

    darknet_train_set: Optional[
        AnnotationHandlerWriteOutputDarknetAnnotationConfig
    ] = related.ChildField(
        cls=AnnotationHandlerWriteOutputDarknetAnnotationConfig,
        required=False,
        default=None,
    )
    darknet_annotation: AnnotationHandlerWriteOutputDarknetAnnotationConfig = related.ChildField(
        cls=AnnotationHandlerWriteOutputDarknetAnnotationConfig,
        required=False,
        default=None,
    )
    use_difficult: bool = related.BooleanField(default=True)
    use_occluded: bool = related.BooleanField(default=True)

    bbox_snippets: Optional[AnnotationHandlerWriteOutputBoxSnippetsConfig] = related.ChildField(
        AnnotationHandlerWriteOutputBoxSnippetsConfig, required=False, default=None
    )

    random_state: Optional[int] = related.ChildField(cls=int, default=None, required=False)


@define
class AnnotationHandlerConfig(BaseConfigClass):
    """Class for parsing information from yaml in respective hierarchy"""

    __related_strict__ = True

    class_mapping: Optional[ClassMappingConfig] = related.ChildField(
        cls=ClassMappingConfig, required=False, default=None
    )

    pascal_voc_input_data: List[AnnotationHandlerPASCALVOCInputDataConfig] = related.SequenceField(
        AnnotationHandlerPASCALVOCInputDataConfig, default=[]
    )

    coco_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = related.SequenceField(
        AnnotationHandlerSingleFileInputDataConfig, default=[]
    )

    cvat_input_data: List[AnnotationHandlerSingleFileInputDataConfig] = related.SequenceField(
        AnnotationHandlerSingleFileInputDataConfig, default=[]
    )

    label_studio_input_data: List[
        AnnotationHandlerLabelStudioInputDataConfig
    ] = related.SequenceField(AnnotationHandlerLabelStudioInputDataConfig, default=[])

    write_output: Optional[AnnotationHandlerWriteOutputConfig] = related.ChildField(
        AnnotationHandlerWriteOutputConfig, required=False, default=None
    )

    reduction_class_mapping: Optional[ReductionMappingConfig] = related.ChildField(
        cls=ReductionMappingConfig, required=False, default=None
    )

    mot_input_data: List[AnnotationHandlerMOTInputDataConfig] = related.SequenceField(
        AnnotationHandlerMOTInputDataConfig, default=[]
    )



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for building CSV formatted annotations."""
import logging
import os
from typing import List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_builder import AnnotationBuilder
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.exceptions import ClassMappingNotFoundError
from mlcvzoo_base.configuration.structs import AnnotationFileFormats, ImageFileFormats
from mlcvzoo_base.data_preparation.utils import ensure_abspath, replace_index_by_dir
from mlcvzoo_base.utils.implicit_path_replacements import ImplicitReplacement

logger = logging.getLogger(__name__)


class CSVAnnotationBuilder(AnnotationBuilder):
    """
    Super class for defining the methods that are needed to build a BaseAnnotation
    object from a CSV type XML file.
    """

    def __init__(
        self,
        csv_line: str,
        mapper: AnnotationClassMapper,
    ) -> None:
        AnnotationBuilder.__init__(self)

        self.csv_line: str = csv_line
        self.mapper: AnnotationClassMapper = mapper

    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        if not os.path.isdir(image_dir):
            raise ValueError(
                f"image_dir '{image_dir}' does not exist. Please provide a valid directory"
            )

        (csv_annotation_path, csv_image_path, image_shape, bounding_boxes,) = self.__init_from_csv(
            image_dir=image_dir,
            annotation_dir=annotation_dir,
        )

        annotation = BaseAnnotation(
            image_path=csv_image_path,
            annotation_path=csv_annotation_path,
            image_shape=image_shape,
            classifications=[],
            bounding_boxes=bounding_boxes,
            segmentations=[],
            image_dir=image_dir,
            annotation_dir=annotation_dir,
            replacement_string=replacement_string,
        )

        try:
            AnnotationBuilder._check_annotation(annotation=annotation)
        except ValueError as value_error:
            logger.exception(
                f"{value_error}, in a future version, the whole annotation will be skipped!"
            )

        annotation = replace_index_by_dir(
            annotation=annotation,
        )

        annotation = ensure_abspath(annotation=annotation)

        return annotation

    def __init_from_csv(
        self,
        image_dir: str,
        annotation_dir: str,
    ) -> Tuple[str, str, Tuple[int, int], List[BoundingBox]]:
        """

        Args:
            image_dir: directory where the image lives
            annotation_dir: directory of the image's annotation

        Returns: path to the annotation,
                path to the annotated image,
                shape of the annotated image and
                a list of annotated bounding_boxes

        """

        bounding_boxes, image_path, image_shape = self.__from_csv_entry()

        file_path = CSVAnnotationBuilder._get_annotation_path(
            image_path, image_dir=image_dir, annotation_dir=annotation_dir
        )

        return file_path, image_path, image_shape, bounding_boxes

    def __from_csv_entry(self) -> Tuple[List[BoundingBox], str, Tuple[int, int]]:
        """
        Parses an image's annotation from a CSV entry (row)

        Returns: a list of annotated bounding_boxes,
            the file path of the image, if available and
            the shape of the annotated image

        """

        line_split = self.csv_line.split(" ")

        image_path = line_split[0]
        image_shape = (int(line_split[1]), int(line_split[2]))

        bounding_boxes: List[BoundingBox] = list()

        for index in range(3, len(line_split)):
            annotation_split = line_split[index].split(",")

            annotation_class_name = annotation_split[4]

            try:
                # map the parsed "class_name" according to the mapping defined in the mapper class
                class_name = self.mapper.map_annotation_class_name_to_model_class_name(
                    class_name=annotation_class_name
                )

                class_id = self.mapper.map_annotation_class_name_to_model_class_id(
                    class_name=annotation_class_name
                )
            except ClassMappingNotFoundError:
                logger.warning(
                    "Could not find a valid class-mapping for class-name '%s'. "
                    "BndBox will be skipped, csv-line-split = '%s'"
                    % (annotation_class_name, line_split)
                )
                continue

            bounding_boxes.append(
                BoundingBox(
                    box=Box(
                        xmin=int(annotation_split[0]),
                        ymin=int(annotation_split[1]),
                        xmax=int(annotation_split[2]),
                        ymax=int(annotation_split[3]),
                    ),
                    class_identifier=ClassIdentifier(class_id=class_id, class_name=class_name),
                    model_class_identifier=ClassIdentifier(
                        class_id=class_id, class_name=class_name
                    ),
                    difficult=False,
                    occluded=False,
                    content="",
                    score=1.0,
                )
            )

        return bounding_boxes, image_path, image_shape

    @staticmethod
    def _get_annotation_path(
        image_path: str,
        image_dir: str,
        annotation_dir: str,
        annotation_encoding: str = AnnotationFileFormats.XML,
    ) -> str:
        """
        Builds the annotation path to an image, assuming images and annotation live in
        similarly structured directories.

        Args:
            image_path: path to the image
            image_dir: directory where the image lives
            annotation_dir: directory of the image's annotation
            annotation_encoding: format of the annotation (one of AnnotationFileFormats)

        Returns: the path to the image's annotation

        """

        annotation_path: str = ImplicitReplacement.replace_directory_in_path(
            file_path=image_path,
            replacement_key=image_dir,
            replacement_value=annotation_dir,
            how=ImplicitReplacement.FIRST,
        )
        annotation_path = annotation_path.replace(ImageFileFormats.JPEG, annotation_encoding)

        annotation_path = annotation_path.replace(ImageFileFormats.PNG, annotation_encoding)

        return annotation_path



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for enumerating options of formats"""
from enum import Enum

from mlcvzoo_base.configuration.structs import BaseType


class CSVOutputStringFormats(BaseType):
    BASE = "BASE"
    YOLO = "YOLO"


class MOTChallengeFormats(Enum):
    MOT15 = "mot15"
    MOT1617 = "mot1617"
    MOT20 = "mot20"



# Copyright 2022 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for evaluating currently trained model checkpoints"""

import logging
from typing import Dict, List, Optional

from tqdm import tqdm

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_class_mapper import AnnotationClassMapper
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.model import ConfigurationType, DataType, ObjectDetectionModel
from mlcvzoo_base.evaluation.object_detection.data_classes import (
    ODModelEvaluationMetrics,
)
from mlcvzoo_base.evaluation.object_detection.metrics_computation import (
    MetricsComputation,
)

logger = logging.getLogger(__name__)


def evaluate_with_model(
    gt_annotations: List[BaseAnnotation],
    iou_thresholds: List[float],
    model: ObjectDetectionModel[ConfigurationType, DataType],
) -> ODModelEvaluationMetrics:
    """
    Compute the metric for the given object detection model. The evaluation is performed
    on the basis of the given ground truth annotations.

    Args:
        model: The model that should be evaluated
        gt_annotations: ground truth annotations where the model should be evaluated on
        iou_thresholds: List of thresholds for which a metrics should be computed

    Returns:
        The computed object detection metrics for this model
    """

    predicted_bounding_boxes_list: List[List[BoundingBox]] = []

    process_bar = tqdm(
        gt_annotations,
        desc=f"Compute metrics",
    )

    for index, gt_annotation in enumerate(process_bar):
        # Every ObjectDetectionModel returns bounding boxes
        _, predicted_bounding_boxes = model.predict(data_item=gt_annotation.image_path)
        predicted_bounding_boxes_list.append(predicted_bounding_boxes)

    return MetricsComputation(
        model_specifier=model.unique_name,
        classes_id_dict=None,
        iou_thresholds=iou_thresholds,
        gt_annotations=gt_annotations,
        predicted_bounding_boxes_list=predicted_bounding_boxes_list,
        mapper=model.mapper,
    ).compute_metrics()


def evaluate_with_precomputed_data(
    model_specifier: str,
    classes_id_dict: Optional[Dict[int, str]],
    gt_annotations: List[BaseAnnotation],
    iou_thresholds: List[float],
    predicted_bounding_boxes_list: List[List[BoundingBox]],
    mapper: Optional[AnnotationClassMapper] = None,
) -> ODModelEvaluationMetrics:
    """
    Compute the object detection metrics taking precomputed (predicted) bounding boxes and
    ground truth annotations.

    IMPORTANT REMARK: The index of the lists 'ground_truth_annotations'
                      and 'predicted_bounding_boxes_list' have to match exactly. This means
                      index 0 indicates the ground truth data and predicted bounding boxes
                      for image 0.

    Args:
        model_specifier: A string to indicate with which model the precomputed bounding boxes
                         have been predicted
        classes_id_dict: (DEPRECATED) A dictionary that defines the mapping of class ids to
                         class names for this evaluation
        gt_annotations: The ground truth data as basis for the evaluation
        iou_thresholds: List of thresholds for which a metrics should be computed
        predicted_bounding_boxes_list: A list containing the predicted bounding boxes for each
                                       image of ground truth data
       mapper: An AnnotationClassMapper object that states the mapping of Class-IDs/Class-Names
               to ClassIdentifier(s)

    Returns:
        The computed object detection metrics
    """

    return MetricsComputation(
        model_specifier=model_specifier,
        classes_id_dict=classes_id_dict,
        iou_thresholds=iou_thresholds,
        gt_annotations=gt_annotations,
        predicted_bounding_boxes_list=predicted_bounding_boxes_list,
        mapper=mapper,
    ).compute_metrics()



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

MODEL_CONFIG_ROOT_DIR_PLACEHOLDER = "MODEL_CONFIG_ROOT_DIR"



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for writing annotations to a file"""
from abc import abstractmethod
from typing import List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation


class AnnotationWriter:
    """
    Super class for defining the methods that are needed to write a list of
    annotations into a file.
    Each annotation format e.g. Pascal-VOC, COCO, CVAT-for-images should have
    its own child class.
    """

    @abstractmethod
    def write(self, annotations: List[BaseAnnotation]) -> Optional[str]:
        """
        Write a given list of annotations into a file.

        Args:
            annotations: list of annotations that should be written

        Returns:
            Optionally path of written file

        """
        raise NotImplementedError()



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for building CVAT formatted annotations."""
import logging
import os
from typing import List, Tuple

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_builder import AnnotationBuilder
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.api.data.segmentation import Segmentation
from mlcvzoo_base.data_preparation.utils import ensure_abspath

logger = logging.getLogger(__name__)


class CVATAnnotationBuilder(AnnotationBuilder):
    """
    Super class for defining the methods that are needed to build a BaseAnnotation
    object from a CVAT type XML file.
    """

    def __init__(
        self,
        image_shape: Tuple[int, int],
        input_classifications: List[Classification],
        input_bounding_boxes: List[BoundingBox],
        input_segmentations: List[Segmentation],
    ) -> None:
        AnnotationBuilder.__init__(self)

        self.image_shape = image_shape
        self.input_classifications = input_classifications
        self.input_bounding_boxes = input_bounding_boxes
        self.input_segmentations = input_segmentations

    def build(
        self,
        image_path: str,
        annotation_path: str,
        image_dir: str,
        annotation_dir: str,
        replacement_string: str,
    ) -> BaseAnnotation:
        annotation = BaseAnnotation(
            image_path=os.path.join(image_dir, image_path),
            annotation_path=os.path.join(annotation_dir, annotation_path),
            image_shape=self.image_shape,
            classifications=self.input_classifications,
            bounding_boxes=self.input_bounding_boxes,
            segmentations=self.input_segmentations,
            image_dir=image_dir,
            annotation_dir=annotation_dir,
            replacement_string=replacement_string,
        )

        try:
            AnnotationBuilder._check_annotation(annotation=annotation)
        except ValueError as value_error:
            logger.exception(
                f"{value_error}, in a future version, the whole annotation will be skipped!"
            )
        annotation = ensure_abspath(annotation=annotation)

        return annotation



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module to provide a registry for model and configuration constructors"""

import copy
import inspect
import logging
from typing import Any, Dict, List, Optional, Type

from config_builder import BaseConfigClass

from mlcvzoo_base.api.model import ConfigurationType, DataType, Model, PredictionType
from mlcvzoo_base.api.registry import MLCVZooRegistry
from mlcvzoo_base.configuration.config_registry import ConfigRegistry
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.configuration.replacement_config import (
    STRING_REPLACEMENT_MAP_KEY,
    ReplacementConfig,
)
from mlcvzoo_base.models.read_from_file.configuration import ReadFromFileConfig
from mlcvzoo_base.models.read_from_file.model import (
    ReadFromFileClassificationModel,
    ReadFromFileObjectDetectionModel,
    ReadFromFileSegmentationModel,
)

logger = logging.getLogger(__name__)


class ModelRegistry(MLCVZooRegistry[Type[Model[PredictionType, ConfigurationType, DataType]]]):
    """
    Class to provide a registry for model and configuration constructors
    """

    def __init__(self) -> None:
        MLCVZooRegistry.__init__(self)

        # =================================================
        # Init and fill ConfigRegistry
        self._config_registry = ConfigRegistry()

        self._config_registry.register_module(
            module_type_name="read_from_file_config", module_constructor=ReadFromFileConfig
        )

        self._config_registry.register_external_module(
            module_type_name="yolox_config",
            module_constructor="YOLOXConfig",
            package_name="mlcvzoo_yolox.configuration",
        )

        self._config_registry.register_external_module(
            module_type_name="tf_classification_xception_config",
            module_constructor="XceptionConfig",
            package_name="mlcvzoo_tf_classification.xception.configuration",
        )

        self._config_registry.register_external_module(
            module_type_name="tf_classification_custom_block_config",
            module_constructor="CustomBlockConfig",
            package_name="mlcvzoo_tf_classification.custom_block.configuration",
        )

        self._config_registry.register_external_module(
            module_type_name="mmocr_config",
            module_constructor="MMOCRConfig",
            package_name="mlcvzoo_mmocr.configuration",
        )

        self._config_registry.register_external_module(
            module_type_name="mmdet_config",
            module_constructor="MMDetectionConfig",
            package_name="mlcvzoo_mmdetection.configuration",
        )
        self._config_registry.register_external_module(
            module_type_name="darknet_config",
            module_constructor="DarknetConfig",
            package_name="mlcvzoo_darknet.configuration",
        )

        # =================================================
        # Fill ModelRegistry
        self.register_model(
            model_type_name="read_from_file_classification",
            model_constructor=ReadFromFileClassificationModel,
        )
        self.register_model(
            model_type_name="read_from_file_object_detection",
            model_constructor=ReadFromFileObjectDetectionModel,
        )
        self.register_model(
            model_type_name="read_from_file_segmentation",
            model_constructor=ReadFromFileSegmentationModel,
        )
        self.register_external_model(
            model_type_name="yolox",
            model_constructor="YOLOXModel",
            package_name="mlcvzoo_yolox.model",
        )
        # TODO: Add deprecation?
        self.register_external_model(
            model_type_name="yolov4_darknet",
            model_constructor="DarknetDetectionModel",
            package_name="mlcvzoo_darknet.model",
        )
        self.register_external_model(
            model_type_name="darknet_object_detection",
            model_constructor="DarknetDetectionModel",
            package_name="mlcvzoo_darknet.model",
        )
        self.register_external_model(
            model_type_name="mmdetection_object_detection",
            model_constructor="MMObjectDetectionModel",
            package_name="mlcvzoo_mmdetection.object_detection_model",
        )
        self.register_external_model(
            model_type_name="mmocr_text_detection",
            model_constructor="MMOCRTextDetectionModel",
            package_name="mlcvzoo_mmocr.text_detection_model",
        )
        self.register_external_model(
            model_type_name="mmocr_text_recognition",
            model_constructor="MMOCRTextRecognitionModel",
            package_name="mlcvzoo_mmocr.text_recognition_model",
        )
        self.register_external_model(
            model_type_name="tf_classification_custom_block",
            model_constructor="CustomBlockModel",
            package_name="mlcvzoo_tf_classification.custom_block.model",
        )
        self.register_external_model(
            model_type_name="tf_classification_xception",
            model_constructor="XceptionModel",
            package_name="mlcvzoo_tf_classification.xception.model",
        )

        self._model_to_config_dict: Dict[str, str] = {
            "read_from_file_classification": "read_from_file_config",
            "read_from_file_object_detection": "read_from_file_config",
            "read_from_file_segmentation": "read_from_file_config",
            "yolox": "yolox_config",
            "yolov4_darknet": "darknet_config",
            "darknet_object_detection": "darknet_config",
            "mmdetection_object_detection": "mmdet_config",
            "mmocr_text_detection": "mmocr_config",
            "mmocr_text_recognition": "mmocr_config",
            "tf_classification_custom_block": "tf_classification_custom_block_config",
            "tf_classification_xception": "tf_classification_xception_config",
        }

    @property
    def model_registry(
        self,
    ) -> Dict[str, Type[Model[PredictionType, ConfigurationType, DataType]]]:
        return self._registry

    @property
    def config_registry(self) -> Dict[str, Type[BaseConfigClass]]:
        return self._config_registry._registry

    def determine_config_class(self, model_type_name: str) -> Type[BaseConfigClass]:
        return self.config_registry[self._model_to_config_dict[model_type_name]]

    def determine_config_class_name(self, model_type_name: str) -> str:
        return self._model_to_config_dict[model_type_name]

    def register_external_model(
        self, model_type_name: str, model_constructor: str, package_name: str
    ) -> None:
        """
        Register an external model
        Args:
            model_type_name: name of the model to register
            model_constructor: name of the constructor of the model to register
            package_name: the full package to import to call the constructor

        Returns:
            Nothing
        """
        self.register_external_module(
            module_type_name=model_type_name,
            module_constructor=model_constructor,
            package_name=package_name,
        )

    def get_registered_models(
        self,
    ) -> Dict[str, Type[Model[PredictionType, ConfigurationType, DataType]]]:
        return copy.deepcopy(self._registry)

    def get_model_type(
        self, class_type: str
    ) -> Optional[Type[Model[PredictionType, ConfigurationType, DataType]]]:
        if class_type in self._registry:
            return self._registry[class_type]

        return None

    def register_model(
        self, model_type_name: str, model_constructor: Any, force: bool = False
    ) -> None:
        self.register_module(
            module_type_name=model_type_name, module_constructor=model_constructor, force=force
        )

    def init_model(
        self,
        model_config: ModelConfig,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> Model[PredictionType, ConfigurationType, DataType]:
        """
        Generic method for instantiating any model that is registered in the model-registry

        Args:
            model_config: The model configuration defining which model should be initialized
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing a configuration file. They can be
                                    understood as variables that can be used to define configs
                                    that are valid across multiple devices.

                                    If no string_replacement_map a default map based on the
                                    ReplacementConfig will be created and used. This allows
                                    to use the attributes of the ReplacementConfig to be
                                    replaced by os environment variables.

        Returns:
            The created model instance
        """

        model_type = self.get_model_type(class_type=model_config.class_type)
        if model_type is not None:
            model: Model[PredictionType, ConfigurationType, DataType]

            init_params: List[Any] = list(inspect.getfullargspec(model_type.__init__).args)
            # We don't need self as parameter in the configuration
            init_params.remove("self")

            if (
                STRING_REPLACEMENT_MAP_KEY not in model_config.constructor_parameters
                and STRING_REPLACEMENT_MAP_KEY in init_params
            ):
                if string_replacement_map is None:
                    string_replacement_map = ReplacementConfig().to_dict()
                model_config.constructor_parameters[
                    STRING_REPLACEMENT_MAP_KEY
                ] = string_replacement_map

            try:
                model = model_type(**model_config.constructor_parameters)  # type: ignore[arg-type]
            except TypeError as e:
                logger.error(
                    "Please provide the parameters " "%s, as specified for %s",
                    init_params,
                    model_type,
                )
                raise e
        else:
            message = (
                f"The model '{model_config.class_type}' is not registered! \n"
                f"The registered models are: {self._registry.keys()}"
            )

            logger.error(message)
            raise ValueError(message)

        return model



# --------------------------------------------------------
# Copyright (c) 2019 Toan Dao Minh(bigkizd)
# Licensed under The MIT License [see LICENSE for details]
# --------------------------------------------------------

"""
This module is used to source out a metric computing utility, which
is taken from https://github.com/toandaominh1997/EfficientDet.Pytorch
"""

from typing import Any

import numpy as np


def compute_overlap(boxes: np.ndarray, query_boxes: np.ndarray) -> Any:  # type: ignore[type-arg]
    """
    Compute overlap between predicted boxes and query_boxes

    Code originally from

    https://github.com/toandaominh1997/EfficientDet.Pytorch/
    blob/fbe56e58c9a2749520303d2d380427e5f01305ba/eval.py#L19-L46

    Args:
        boxes: (N, 4) ndarray of float
        query_boxes: (K, 4) ndarray of float

    Returns
        (N, K) ndarray of overlap between predicted boxes and query_boxes
    """

    assert type(boxes) == np.ndarray
    assert type(query_boxes) == np.ndarray

    area = (query_boxes[:, 2] - query_boxes[:, 0]) * (query_boxes[:, 3] - query_boxes[:, 1])

    iw = np.minimum(np.expand_dims(boxes[:, 2], axis=1), query_boxes[:, 2]) - np.maximum(
        np.expand_dims(boxes[:, 0], 1), query_boxes[:, 0]
    )
    ih = np.minimum(np.expand_dims(boxes[:, 3], axis=1), query_boxes[:, 3]) - np.maximum(
        np.expand_dims(boxes[:, 1], 1), query_boxes[:, 1]
    )

    iw = np.maximum(iw, 0)
    ih = np.maximum(ih, 0)

    ua = (
        np.expand_dims((boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1]), axis=1)
        + area
        - iw * ih
    )

    ua = np.maximum(ua, np.finfo(float).eps)

    intersection = iw * ih

    # normally this is of type np.ndarray: TODO: solve mypy error
    return intersection / ua



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""
Module for storing data classes and complex type definitions that are used in the context of the
mlcvzoo_base.evaluation.object_detection package
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Final, List, Optional

import numpy as np

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.bounding_box import BoundingBox

# 3D List containing of the following:
#   1) One entry per data items / image
#   2) One entry per call of classes
#   3) One entry per annotation Annotations the matches the image (1) and the class (2)
# This is a list of BaseAnnotations in order to carry information about the image
# instead of just bounding boxes
EVALUATION_LIST_TYPE = List[List[List[BaseAnnotation]]]  # pylint: disable=invalid-name

CONFUSION_MATRIX_TYPE = List[List[int]]
CONFUSION_MATRIX_DICT_TYPE = Dict[str, Dict[str, int]]  # pylint: disable=invalid-name

DEFAULT_INT_VALUE: Final[int] = 0
DEFAULT_FLOAT_VALUE: Final[float] = 0.0


# NOTE: Since this are the main Object Detection metrics, it's okay to have more instance
#       attributes
@dataclass
class ODMetrics:  # pylint: disable=too-many-instance-attributes
    """
    Dataclass for storing the main metrics that are computed for object detection algorithms
    """

    # TODO: rename to lower case, disable pylint only temporary!
    TP: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    FP: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    FN: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name
    PR: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    RC: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    F1: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    AP: float = DEFAULT_FLOAT_VALUE  # pylint: disable=invalid-name
    COUNT: int = DEFAULT_INT_VALUE  # pylint: disable=invalid-name

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> ODMetrics:
        return ODMetrics(**input_dict)

    def __repr__(self):  # type: ignore
        return (
            f"TP: {self.TP}, "
            f"FP: {self.FP}, "
            f"FN: {self.FN}, "
            f"PR: {self.PR}, "
            f"RC: {self.RC}, "
            f"F1: {self.F1}, "
            f"AP: {self.AP}, "
            f"COUNT: {self.COUNT}"
        )

    def __str__(self):  # type: ignore
        return self.__repr__()


@dataclass
class MetricImageInfo:
    """
    Dataclass to store information about false positives and false negatives in the
    form of BaseAnnotation objects. This to have a exact relation between an image
    and the according false positive / false negative bounding boxes. The ground
    truth data is added to be able to visualize the expected bounding boxes.
    """

    ground_truth_annotation: Optional[BaseAnnotation] = None
    false_negative_annotation: Optional[BaseAnnotation] = None
    false_positive_annotation: Optional[BaseAnnotation] = None
    false_negative_matched_false_positive_annotation: Optional[BaseAnnotation] = None


# 1st key: Class Identifier as string
# 2nd key: Image Path
# 2nd value: The MetricImageInfo for this class name and image path
METRIC_IMAGE_INFO_TYPE = Dict[str, Dict[str, MetricImageInfo]]  # pylint: disable=invalid-name


# 1st key: iou-threshold
# 2nd key: Type of the size of the bounding-box => Any of BBoxSizeTypes.BBOX_SIZE_TYPE
# 3rd key: Class Identifier as string
# value: The computed metrics of type ODMetrics
METRIC_DICT_TYPE = Dict[float, Dict[str, Dict[str, ODMetrics]]]  # pylint: disable=invalid-name


def build_metric_dict_from_dict(
    input_dict: Dict[str, Dict[str, Dict[str, Dict[str, Any]]]]
) -> METRIC_DICT_TYPE:
    metric_dict: METRIC_DICT_TYPE = {}

    for key_0, value_0 in input_dict.items():
        metric_dict[float(key_0)] = {}
        for key_1, value_1 in value_0.items():
            metric_dict[float(key_0)][key_1] = {}
            for key_2, od_metrics_dict in value_1.items():
                metric_dict[float(key_0)][key_1][key_2] = ODMetrics.from_dict(
                    input_dict=od_metrics_dict
                )

    return metric_dict


@dataclass
class ODModelEvaluationMetrics:
    """
    Dataclass for storing the output of an object detection evaluation.
    The metrics_dict stores the actual computed metrics, while the metrics_image_info_dict
    stores debugging information to be able to analyze false positives and false negatives.

    The model_specifier indicates for which model the metrics have been computed.
    """

    model_specifier: str
    metrics_dict: METRIC_DICT_TYPE = field(default_factory=lambda: {})
    metrics_image_info_dict: METRIC_IMAGE_INFO_TYPE = field(default_factory=lambda: {})


@dataclass
class ODEvaluationComputingData:
    """
    Dataclass for storing data structures that are needed to computed object detection metrics
    """

    # 1st key: Type of the size of the bounding-box => Any of BBoxSizeTypes.BBOX_SIZE_TYPE
    # 2nd key: Class Identifier as string
    # value: The number of ground truth boxes for the combination of keys
    gt_counter_dict: Dict[str, Dict[str, int]] = field(default_factory=lambda: {})

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of BBoxSizeTypes.BBOX_SIZE_TYPE
    # Cumulative array indicating the false positives of the dataset
    false_positives_dict: Dict[float, Dict[str, np.ndarray]] = field(  # type: ignore[type-arg]
        default_factory=lambda: {}
    )

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of BBoxSizeTypes.BBOX_SIZE_TYPE
    # Cumulative array indicating the true positives of the dataset
    true_positives_dict: Dict[float, Dict[str, np.ndarray]] = field(  # type: ignore[type-arg]
        default_factory=lambda: {}
    )

    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of BBoxSizeTypes.BBOX_SIZE_TYPE
    # value: Array indicating the score for each data item
    scores: Dict[float, Dict[str, np.ndarray]] = field(  # type: ignore[type-arg]
        default_factory=lambda: {}
    )

    # key: iou-threshold
    detected_annotations: Dict[float, List[BoundingBox]] = field(default_factory=lambda: {})

    # TODO: Is this needed?! => Remove at next major version update
    # 1st key: iou-threshold
    # 2nd key: Type of the size of the bounding-box => Any of BBoxSizeTypes.BBOX_SIZE_TYPE
    # List indicating the AP metrics that have been computed correctly
    valid_precisions: Dict[float, Dict[str, List[float]]] = field(default_factory=lambda: {})



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Dataclass that holds attributes of Annotation objects"""


class AnnotationAttributes:
    """Class for describing different annotation attributes"""

    def __init__(
        self,
        difficult: bool,
        occluded: bool,
        content: str,
    ):
        self.__difficult: bool = difficult
        self.__occluded: bool = occluded
        self.__content: str = content

    @property
    def difficult(self) -> bool:
        return self.__difficult

    @property
    def occluded(self) -> bool:
        return self.__occluded

    @property
    def content(self) -> str:
        return self.__content



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module that holds classes for different types of basic dataset information."""
import os
from dataclasses import dataclass, field
from typing import Dict

from dataclasses_json import dataclass_json

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.utils.implicit_path_replacements import ImplicitReplacement


@dataclass_json
@dataclass
class ClassesCountInfo:
    count: str
    count_s: str
    count_m: str
    count_l: str


@dataclass_json
@dataclass
class DatasetDirectoryInfo:
    """Class holds information about a specific directory"""

    file_count: int
    classes_count_info: Dict[str, ClassesCountInfo] = field(default_factory=dict)


@dataclass_json
@dataclass
class BaseDatasetInfo:
    """Class holds information about a specific dataset"""

    classes_path: str
    image_count: float

    directory_info: Dict[str, DatasetDirectoryInfo] = field(default_factory=dict)

    base_file: str = ""

    def update(self, annotation: BaseAnnotation) -> None:
        """
        Updates the class's directory information with the given annotation.

        Args:
            annotation: BaseAnnotation object

        """

        self.image_count += 1

        # get relative directory of the annotation
        relative_dirname = os.path.dirname(
            ImplicitReplacement.replace_directory_in_path(
                file_path=annotation.annotation_path,
                replacement_key=annotation.annotation_dir,
                replacement_value="",
            )
        )

        # Init info dict
        if relative_dirname not in self.directory_info:
            self.directory_info[relative_dirname] = DatasetDirectoryInfo(file_count=1)
        else:
            self.directory_info[relative_dirname].file_count += 1

        # Fill info dict with statistics about the annotation
        # TODO:
        # for bounding_box in annotation.get_bounding_boxes(include_segmentations=True):
        #     bbox_height = (bounding_box.box.ymax - bounding_box.box.ymin) / annotation.image_shape[0] * 480
        #     bbox_width = (bounding_box.box.xmax - bounding_box.box.xmin) / annotation.image_shape[1] * 640
        #
        #     bbox_sqrt_area = math.sqrt(bbox_height * bbox_width)
        #
        #     if bounding_box.class_name in info_dict[relative_dirname]:
        #         info_dict[relative_dirname][bounding_box.class_name]["count"] += 1
        #
        #         if bbox_sqrt_area <= 32:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_s"] += 1
        #
        #         if 32 < bbox_sqrt_area <= 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_m"] += 1
        #
        #         if bbox_sqrt_area > 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_l"] += 1
        #     else:
        #         info_dict[relative_dirname][bounding_box.class_name] = dict()
        #
        #         info_dict[relative_dirname][bounding_box.class_name]["count"] = 1
        #
        #         info_dict[relative_dirname][bounding_box.class_name]["count_s"] = 0
        #         info_dict[relative_dirname][bounding_box.class_name]["count_m"] = 0
        #         info_dict[relative_dirname][bounding_box.class_name]["count_l"] = 0
        #
        #         if bbox_sqrt_area <= 32:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_s"] += 1
        #
        #         if 32 < bbox_sqrt_area <= 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_m"] += 1
        #
        #         if bbox_sqrt_area > 96:
        #             info_dict[relative_dirname][bounding_box.class_name]["count_l"] += 1



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

"""Module for writing CVAT formatted annotations."""
import logging
import xml.etree.ElementTree as ET_xml
from typing import List, Optional

from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.data.annotation_writer import AnnotationWriter
from mlcvzoo_base.utils.xml_utils import xml_tree_to_file

logger = logging.getLogger(__name__)


class CVATAnnotationWriter(AnnotationWriter):
    """
    Writer for generating a CVAT 1.1 conform xml file
    out of a list of annotations.
    """

    def __init__(
        self,
        cvat_xml_input_path: str,
        output_file_path: str,
        clean_boxes: bool = False,
        clean_segmentations: bool = False,
        clean_tags: bool = False,
    ):
        self.cvat_xml_input_path = cvat_xml_input_path
        self.output_file_path = output_file_path
        self.clean_boxes = clean_boxes
        self.clean_segmentations = clean_segmentations
        self.clean_tags = clean_tags

    def write(
        self,
        annotations: List[BaseAnnotation],
    ) -> Optional[str]:
        xml_tree = CVATAnnotationWriter.annotations_to_xml_tree(
            annotations=annotations,
            cvat_xml_input_path=self.cvat_xml_input_path,
            clean_boxes=self.clean_boxes,
            clean_segmentations=self.clean_segmentations,
            clean_tags=self.clean_tags,
        )

        # NOTE: generated file will overwrite existing file
        xml_tree_to_file(xml_file_path=self.output_file_path, xml_tree=xml_tree)

        return self.output_file_path

    @staticmethod
    def annotations_to_xml_tree(
        annotations: List[BaseAnnotation],
        cvat_xml_input_path: str,
        clean_boxes: bool = False,
        clean_segmentations: bool = False,
        clean_tags: bool = False,
    ) -> ET_xml.ElementTree:
        """
        Parses given annotations to an XML ElementTree.

        Args:
            annotations: a list of BaseAnnotation objects
            cvat_xml_input_path: string, the path to CVAT annotations in XML format
            clean_boxes: bool, whether to overwrite or add box annotations
            clean_segmentations: bool, whether to overwrite or add segmentation annotations
            clean_tags: bool, whether to overwrite or add tag annotations

        Returns: an XML ElementTree

        """
        tree = ET_xml.parse(source=cvat_xml_input_path)
        root = tree.getroot()

        annotation_dict = {}
        for annotation in annotations:
            annotation_dict[annotation.image_path] = annotation

        images = root.findall("image")
        for image in images:
            image_relative_path = image.attrib["name"]

            # create dictionary of image_path -> annotation,
            #  where the image_path ends with the relative path listed in "name" attribute
            corresponding_annotation_dict = {
                key: val
                for key, val in annotation_dict.items()
                if key.endswith(image_relative_path)
            }

            if len(list(corresponding_annotation_dict.keys())) != 1:
                logger.warning(
                    "No precise entry for image with "
                    "relative path {image_relative_path} in annotations. "
                    "Found %s entries. "
                    "Skipping that image tag in XML.",
                    len(list(corresponding_annotation_dict.keys())),
                )
                continue

            # use corresponding annotation
            image_path = list(corresponding_annotation_dict.keys())[0]
            current_annotation = annotation_dict[image_path]

            # add bbox info
            CVATAnnotationWriter._add_box_subtree(image, current_annotation, clean_boxes)

            # add seg info
            CVATAnnotationWriter._add_segmentation_subtree(
                image, current_annotation, clean_segmentations
            )

            # add tag info
            CVATAnnotationWriter._add_classification_subtree(image, current_annotation, clean_tags)

        return tree

    @staticmethod
    def _add_box_subtree(
        sub_tree: ET_xml.Element, annotation: BaseAnnotation, clean_boxes: bool = False
    ) -> None:
        if clean_boxes:
            # remove existing bounding boxes
            for elem in list(sub_tree):
                if elem.tag == "box":
                    map(elem.remove, list(elem))
                    sub_tree.remove(elem)

        for bbox in annotation.bounding_boxes:
            bbox_dict = dict()

            bbox_dict["xtl"] = str(bbox.box.xmin)
            bbox_dict["ytl"] = str(bbox.box.ymin)
            bbox_dict["xbr"] = str(bbox.box.xmax)
            bbox_dict["ybr"] = str(bbox.box.ymax)

            bbox_dict["label"] = bbox.class_name
            bbox_dict["source"] = "generated"
            bbox_dict["z_order"] = str(0)

            occluded = bbox.occluded
            occluded_value = 0
            if occluded:
                occluded_value = 1
            bbox_dict["occluded"] = str(occluded_value)

            box_tag = ET_xml.SubElement(sub_tree, "box", attrib=bbox_dict)

            difficult = bbox.difficult
            content = bbox.content

            difficult_tag = ET_xml.SubElement(box_tag, "attribute", attrib={"name": "difficult"})
            if difficult:
                difficult_tag.text = "true"
            else:
                difficult_tag.text = "false"

            occluded_tag = ET_xml.SubElement(box_tag, "attribute", attrib={"name": "occluded"})
            if occluded:
                occluded_tag.text = "true"
            else:
                occluded_tag.text = "false"

            content_tag = ET_xml.SubElement(box_tag, "attribute", attrib={"name": "content"})
            content_tag.text = content

    @staticmethod
    def _add_segmentation_subtree(
        sub_tree: ET_xml.Element,
        annotation: BaseAnnotation,
        clean_segmentations: bool = False,
    ) -> None:
        if clean_segmentations:
            # remove existing polygons
            for elem in list(sub_tree):
                if elem.tag == "polygon":
                    map(elem.remove, list(elem))
                    sub_tree.remove(elem)

        for segmentation in annotation.segmentations:
            segmentation_dict = dict()

            segmentation_dict["points"] = segmentation.to_points_string()

            segmentation_dict["label"] = segmentation.class_name
            segmentation_dict["source"] = "generated"
            segmentation_dict["z_order"] = str(0)

            occluded = segmentation.occluded
            occluded_value = 0
            if occluded:
                occluded_value = 1
            segmentation_dict["occluded"] = str(occluded_value)

            polygon_tag = ET_xml.SubElement(sub_tree, "polygon", attrib=segmentation_dict)

            difficult = segmentation.difficult
            content = segmentation.content

            difficult_tag = ET_xml.SubElement(
                polygon_tag, "attribute", attrib={"name": "difficult"}
            )
            if difficult:
                difficult_tag.text = "true"
            else:
                difficult_tag.text = "false"

            occluded_tag = ET_xml.SubElement(polygon_tag, "attribute", attrib={"name": "occluded"})
            if occluded:
                occluded_tag.text = "true"
            else:
                occluded_tag.text = "false"

            content_tag = ET_xml.SubElement(polygon_tag, "attribute", attrib={"name": "content"})
            content_tag.text = content

    @staticmethod
    def _add_classification_subtree(
        sub_tree: ET_xml.Element, annotation: BaseAnnotation, clean_tags: bool = False
    ) -> None:
        if clean_tags:
            # remove existing classification tags
            for elem in list(sub_tree):
                if elem.tag == "tag":
                    map(elem.remove, list(elem))
                    sub_tree.remove(elem)

        for classification in annotation.classifications:
            classification_dict = dict()
            classification_dict["label"] = classification.class_name
            classification_dict["source"] = "generated"
            polygon_tag = ET_xml.SubElement(sub_tree, "tag", attrib=classification_dict)
            difficult_tag = ET_xml.SubElement(
                polygon_tag, "attribute", attrib={"name": "difficult"}
            )
            difficult_tag.text = "false"



# Copyright 2021 Open Logistics Foundation
#
# Licensed under the Open Logistics License 1.0.
# For details on the licensing terms, see the LICENSE file.

""" Class for Segmentation (polygon areas) annotations"""
from __future__ import annotations

import math
from typing import Any, Dict, List, Optional, Tuple, Type

import cv2
import numpy as np

from mlcvzoo_base.api.data.annotation_attributes import AnnotationAttributes
from mlcvzoo_base.api.data.bounding_box import BoundingBox
from mlcvzoo_base.api.data.box import Box
from mlcvzoo_base.api.data.class_identifier import ClassIdentifier
from mlcvzoo_base.api.data.classification import Classification
from mlcvzoo_base.configuration.structs import ObjectDetectionBBoxFormats
from mlcvzoo_base.third_party.imutils.perspective import order_points

PointType = Tuple[float, float]  # a point defined as (x, y)
PolygonType = List[PointType]


class Segmentation(AnnotationAttributes, Classification):
    """
    A class for defining the data object consumed by Segmentation models.
    It is mainly described by its polygon (list of 2D points) attribute.
    The polygon captures an area of the image, that is defined by the
    inner of the linear connected points of the polygon and is associated
    with a certain class.
    In addition, a segmentation CAN have an attribute "box" that defines the rectangle
    which encloses the area of the polygon.
    """

    def __init__(
        self,
        polygon: PolygonType,
        class_identifier: ClassIdentifier,
        score: float,
        difficult: bool,
        occluded: bool,
        content: str,
        box: Optional[Box] = None,
        model_class_identifier: Optional[ClassIdentifier] = None,
    ):
        Classification.__init__(
            self,
            class_identifier=class_identifier,
            model_class_identifier=model_class_identifier,
            score=score,
        )
        AnnotationAttributes.__init__(
            self, difficult=difficult, occluded=occluded, content=content
        )
        self.__polygon: PolygonType = polygon
        self.__box: Optional[Box] = box

    @property
    def polygon(self) -> PolygonType:
        return self.__polygon

    @property
    def box(self) -> Optional[Box]:
        return self.__box

    def __eq__(self, other: Segmentation):  # type: ignore
        # NOTE: Since floats may very for different systems, don't check the score for equality,
        #       but allow it to be in a reasonable range
        basic_compare: bool = (
            self.box == other.box
            and self.class_identifier.class_id == other.class_identifier.class_id
            and self.class_identifier.class_name == other.class_identifier.class_name
            and self.model_class_identifier.class_id == other.model_class_identifier.class_id
            and self.model_class_identifier.class_name == other.model_class_identifier.class_name
            and self.occluded == other.occluded
            and self.difficult == other.difficult
            and self.content == other.content
            and math.isclose(a=self.score, b=other.score, abs_tol=0.005)
        )

        polygon_compare = True

        if len(self.polygon) == len(other.polygon):
            for polygon_point, other_polygon_point in zip(self.polygon, other.polygon):
                point_equal = (
                    polygon_point[0] == other_polygon_point[0]
                    and polygon_point[1] == other_polygon_point[1]
                )

                if point_equal:
                    polygon_compare = polygon_compare and point_equal
                else:
                    polygon_compare = False
                    break
        else:
            polygon_compare = False

        return basic_compare and polygon_compare

    def __repr__(self):  # type: ignore
        return (
            f"Segmentation("
            f"class-id={self.class_id}, "
            f"class-name={self.class_name}: "
            f"model-class-id={self.model_class_identifier.class_id}, "
            f"model-class-name={self.model_class_identifier.class_name}: "
            f"Box={self.box}, "
            f"Polygon={self.polygon}, "
            f"score={self.score}, "
            f"difficult={self.difficult}, "
            f"occluded={self.occluded}, "
            f"content='{self.content}')"
        )

    def to_dict(self, raw_type: bool = False, reduced: bool = False) -> Dict[str, Any]:
        if self.box:
            _box = self.box if raw_type else self.box.to_dict()
        else:
            _box = None

        if reduced:
            return {
                "box": _box,
                "polygon": self.polygon,
                "class_id": self.class_id,
                "class_name": self.class_name,
                "model_class_id": self.model_class_identifier.class_id,
                "model_class_name": self.model_class_identifier.class_name,
                "score": self.score,
            }
        else:
            return {
                "box": _box,
                "polygon": self.polygon,
                "class_identifier": self.class_identifier
                if raw_type
                else self.class_identifier.to_dict(),
                "model_class_identifier": self.model_class_identifier
                if raw_type
                else self.model_class_identifier.to_dict(),
                "score": self.score,
                "difficult": self.difficult,
                "occluded": self.occluded,
                "content": self.content,
            }

    @staticmethod
    def from_dict(input_dict: Dict[str, Any], reduced: bool = False) -> Segmentation:
        # fmt: off
        if reduced:
            return Segmentation(**{
                "box": Box(**input_dict["box"]),
                "polygon": input_dict["polygon"],
                "class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["class_id"],
                    "class_name": input_dict["class_name"],
                }),
                "model_class_identifier": ClassIdentifier(**{
                    "class_id": input_dict["model_class_id"],
                    "class_name": input_dict["model_class_name"],
                }),
                "score": input_dict["score"],
                "difficult": False,
                "occluded": False,
                "content": "",
            })
        else:
            return Segmentation(**{
                "box": Box(**input_dict["box"]),
                "polygon": input_dict["polygon"],
                "class_identifier": ClassIdentifier(
                    **input_dict["class_identifier"]
                ),
                "model_class_identifier": ClassIdentifier(
                    **input_dict["model_class_identifier"]
                ),
                "score": input_dict["score"],
                "difficult": input_dict["difficult"],
                "occluded": input_dict["occluded"],
                "content": input_dict["content"],
            })
        # fmt: on

    def to_json(self) -> Any:
        return self.to_dict(raw_type=False)

    def to_bounding_box(self, image_shape: Optional[Tuple[int, int]] = None) -> BoundingBox:
        """
        Manually build the bounding box object which covers the rectangular area
        that is defined by the polygon attribute

        Returns: a Bounding Box version of Segmentation

        """

        return BoundingBox(
            class_identifier=self.class_identifier,
            model_class_identifier=self.model_class_identifier,
            score=self.score,
            box=self.get_box(image_shape=image_shape),
            difficult=self.difficult,
            occluded=self.occluded,
            content=self.content,
        )

    def to_list(self, dst_type: Type[Any] = int) -> List[List[float]]:
        """
        Produce an encapsulated list of polygon points
        [[x0, y0, x1, y1, ..., xn, yn]]

        Args:
            dst_type: destination type to transform the data to

        Returns:
            the produced list
        """

        polygon_list: List[float] = []

        for polygon_tuple in self.polygon:
            polygon_list.append(dst_type(polygon_tuple[0]))
            polygon_list.append(dst_type(polygon_tuple[1]))

        return [polygon_list]

    def to_points_string(self) -> str:
        """

        Returns: Sting, concatenation of the coordinates of polygon attribute

        """

        points_string = ""
        for i, point in enumerate(self.polygon):
            if i < len(self.polygon) - 1:
                points_string = (
                    points_string + str(float(point[0])) + "," + str(float(point[1])) + ";"
                )
            else:
                points_string = points_string + str(float(point[0])) + "," + str(float(point[1]))

        return points_string

    def get_box(self, image_shape: Optional[Tuple[int, int]] = None) -> Box:
        """

        Returns: Box object that holds information about bounding box coordinates

        """
        if self.box is not None:
            return self.box
        else:
            return Segmentation.polygon_to_box(polygon=self.polygon, image_shape=image_shape)

    @staticmethod
    def polygon_to_box(polygon: PolygonType, image_shape: Optional[Tuple[int, int]] = None) -> Box:
        """
        Produce a box object that covers the minimal rectangular area which is
        covered by the given polygon

        Args:
            polygon: PolygonType, a list of points that form the polygon
            image_shape: shape of the image for which the bounding box should
                         be created

        Returns:
            A Box that holds the coordinates of the bounding rectangle of the polygon
        """
        return Box.init_format_based(
            box_format=ObjectDetectionBBoxFormats.XYWH,
            box_list=(cv2.boundingRect(np.array(polygon).astype(int))),
            src_shape=image_shape,
        )

    def to_rect_polygon(self) -> None:
        """
        Transform a polygon

        Returns:
            A PolygonType (an ordered list of PointType) that define rectangle
            around the original polygon

        """
        np_sorted_rect_polygon = Segmentation.__polygon_to_sorted_rect_polygon(
            np_polygon=np.array(self.polygon)
        )

        # TODO: limit point coordinates to image-size -1.0 ?
        self.__polygon = [(point[0], point[1]) for point in np_sorted_rect_polygon]

    @staticmethod
    def __polygon_to_sorted_rect_polygon(np_polygon: np.ndarray) -> np.ndarray:  # type: ignore[type-arg]
        """
        Calculates the minimum area around a given numpy array of polygons.
        The input format is nx2 =>  [(x, y), (x, y), ...].
        The output will be a numpy array out of 4 polygon-points => 4x2,
        which define the rect around the original polygon
        """

        min_area_rect = cv2.minAreaRect(np_polygon.astype(int))
        box = order_points(polygon=cv2.boxPoints(min_area_rect), sort_by_euclidean=False)
        return np.intp(box).astype(np.float32)  # type: ignore[return-value]

    def copy_segmentation(self, class_identifier: ClassIdentifier) -> Segmentation:
        return Segmentation(
            polygon=self.polygon,
            class_identifier=class_identifier,
            score=self.score,
            difficult=self.difficult,
            occluded=self.occluded,
            content=self.content,
            box=self.box,
            model_class_identifier=self.model_class_identifier,
        )
